#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  shiftcrypt.py
#
#  Copyright 2018  <@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import sys

from b2bTools.general.Io import B2bIo
from b2bTools.singleSeq.EFoldMine.Predictor import EFoldMine
import warnings
warnings.filterwarnings("ignore")
import argparse
from b2bTools.singleSeq.Agmata.sources.agmata_source import agmata

def getAgmataPredictions(seqs, efm = None):

	# Best cutoff for aggregating, or not
	threshold = 0.015563146

	model=agmata(verbose=-1)

	try:
		model.load()
	except:
		raise
		print('Error loading the model. Please double check you installed all the dependencies. If everything is correct, please report the bug to wim.vranken@vub.be')
		return

	if not efm:
		print('Running EFoldMine')
		efm = EFoldMine()
		efm.predictSeqs(seqs)

	try:
		results = model.predict(seqList=seqs,dmPredictions=efm.allPredictions)

		for (proteinId, sequence) in seqs:
			# Loop over sequence, set category
			efm.allPredictions[proteinId]['agmata'] = []
			for i in range(len(results[proteinId])):
				aggrclass = 0
				if results[proteinId][i] > threshold:
					aggrclass = 1

				results[proteinId][i] = (sequence[i],results[proteinId][i],aggrclass)
				efm.allPredictions[proteinId]['agmata'].append(results[proteinId][i])

	except:
		raise
		print('Error in the prediction. Please double check you installed all the dependencies. If everything is correct, please report the bug to wim.vranken@vub.be')
		return

	return results

def run_agmata(args):
	args=args[1:]
	pa = argparse.ArgumentParser()

	pa.add_argument('-i', '--infile',
						help='the input FASTA file',
						)
	pa.add_argument('-o', '--outfile',
						help='output file',
						default=None)

	parseArgs = pa.parse_args(args)

	b2bio = B2bIo()

	try:
		seqs = b2bio.readFasta(parseArgs.infile)
	except:
		print('Error in FASTA file parsing. Please double check the format and input file name. If everything is correct, please report the bug to wim.vranken@vub.be')
		return

	results = getAgmataPredictions(seqs)

	textOutput = ""

	textOutput+= '#AA AgmataScore BinaryPrediction\n'
	for proteinId in results.keys():
		textOutput+= '# {}\n'.format(proteinId)
		for (aaCode,prediction,predClass) in results[proteinId]:
			textOutput+="{} {:7.2e} {}\n".format(aaCode,prediction,predClass)

	if parseArgs.outfile!=None:
		f = open(parseArgs.outfile, 'w')
		f.write(textOutput)
		f.close()

	else:
		print(textOutput)

	print('\nDONE')

def main(args):
	return 0

if __name__ == '__main__':
	sys.exit(run_agmata(sys.argv))



"""
======================COPYRIGHT/LICENSE START==========================

sequenceIO.py: I/O for nmrStar sequence information saveframes

Copyright (C) 2005-2008 Wim Vranken (European Bioinformatics Institute)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

A copy of this license can be found in ../../../../license/LGPL.license

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

- contact Wim Vranken (wim@ebi.ac.uk)
=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

===========================REFERENCE END===============================
"""

from b2bTools.general.ccpn.format.nmrStar.generalIO import NmrStarFile, NmrStarGenericFile

#from b2bTools.general.ccpn.format.nmrStar.projectIO import MoleculeDbLink

from .constants import bmrbCodeToCcpCode

from b2bTools.general.ccpn.universal.Util import returnInt

#####################
# Class definitions #
#####################

class NmrStarFile(NmrStarFile):

  def initialize(self, version = '2.1.1'):

    self.sequenceFiles = []

    self.files = self.sequenceFiles

    if not self.version:
      self.version = version

    self.DataClassFile = NmrStarSequenceFile

    self.setComponents()

  def setComponentList(self):

    otherSfs = []

    if self.version in ('3.0','3.1'):
      self.saveFrameName = 'assembly'
      otherSfs = ['entity','chem_comp']
    else:
      self.saveFrameName = 'monomeric_polymer'
      otherSfs = ['molecular_system']

    self.components = [self.saveFrameName] + otherSfs

  def read(self,verbose = 0):

    self.readComponent(verbose = verbose)

class NmrStarSequenceFile(NmrStarGenericFile):

  """
  Information on file level.
  """

  def initialize(self,parent,saveFrame = None):

    self.sequences = []
    self.dbLinks = []
    self.bioFunctions = []

    self.saveFrame = saveFrame

    self.molSystemCode = None
    self.details = None
    self.EC_number = None

    self.parent = parent
    self.version = parent.version

    if self.saveFrame:
      self.parseSaveFrame()

  def parseSaveFrame(self):

    if not self.checkVersion():
      return

    if self.version in ('3.0','3.1'):

      self.molSystemCode = self.saveFrame.tags['Name']

      if 'Details' in self.saveFrame.tags:
        self.details = self.saveFrame.tags['Details']

      if 'Enzyme_commission_number' in self.saveFrame.tags:
        self.EC_number = self.saveFrame.tags['Enzyme_commission_number']

      if '_Entity_assembly' in self.saveFrame.tables:

        #
        # First look for original sequence codes if available from coordinate section. Hack for reading combined NMR-STAR files
        #

        if self.version in ('3.1',):

          self.origSeqCodeBySeqId = {}
          self.origSeqInsertCodeBySeqId = {}
          self.atomNamesBySeqId = {}

          if 'conformer_family_coord_set' in self.parent.sfs:
            for coordSf in self.parent.sfs['conformer_family_coord_set']:
              if '_Atom_site' in coordSf.tables:
                coordTableTags = coordSf.tables['_Atom_site'].tags
                for i in range(len(coordTableTags['Model_ID'])):
                  chainId  = coordTableTags['Label_entity_assembly_ID'][i]
                  seqId    = coordTableTags['Label_comp_index_ID'][i]
                  atomName = coordTableTags['Label_atom_ID'][i]

                  if chainId and seqId:
                    residueKey = (chainId,seqId)
                    if not residueKey in self.origSeqCodeBySeqId:
                      if 'Auth_seq_ID' in coordTableTags:
                        seqCode = coordTableTags['Auth_seq_ID'][i]
                    elif 'PDB_residue_no' in coordTableTags:
                        seqCode = coordTableTags['PDB_residue_no'][i]

                    if seqCode != None:
                        self.origSeqCodeBySeqId[residueKey] = returnInt(seqCode)

                    if not residueKey in self.origSeqInsertCodeBySeqId and 'PDB_ins_code' in coordTableTags:
                      seqInsertCode = coordTableTags['PDB_ins_code'][i]
                      if seqInsertCode:
                        self.origSeqInsertCodeBySeqId[residueKey] = seqInsertCode

                    if atomName:
                      if not residueKey in self.atomNamesBySeqId:
                        self.atomNamesBySeqId[residueKey] = []

                      if not atomName in self.atomNamesBySeqId[residueKey]:
                        self.atomNamesBySeqId[residueKey].append(atomName)

        #
        # Now start to read the sequence info
        #

        tableTags = self.saveFrame.tables['_Entity_assembly'].tags

        for i in range(len(tableTags['Entity_ID'])):

          entityName = tableTags['Entity_label'][i]
          entityAssemblyId =  tableTags['ID'][i]

          if self.version in ('3.1',):
            originalChainCode = tableTags['Asym_ID'][i]
            entityId = tableTags['Entity_ID'][i]
            details = tableTags['Details'][i]

            if 'Role' in tableTags:
              role = tableTags['Role'][i]
            else:
              role = None

            if 'Physical_state' in tableTags:
              physState = tableTags['Physical_state'][i]
            else:
              physState = None

            if 'Conformational_isomer' in tableTags:
              confIsomer = tableTags['Conformational_isomer'][i]
            else:
              confIsomer = None

            if 'Chemical_exchange_state' in tableTags:
              chemExchState = tableTags['Chemical_exchange_state'][i]
            else:
              chemExchState = None

            if 'Magnetic_equivalence_group_code' in tableTags:
              magnEquivCode = tableTags['Magnetic_equivalence_group_code'][i]
            else:
              magnEquivCode = None

          else:
            originalChainCode = None
            entityId = None

          for entitySaveFrame in self.parent.sfs['entity']:
            entityFound = False
            if self.version in ('3.0','3.1') and entitySaveFrame.tags['ID'] == entityId:
              entityFound = True
            elif entitySaveFrame.title == entityName[1:]:
              entityFound = True

            if entityFound:

              #waterMoleculeInfo = self.getWaterMoleculeInfo(entityAssemblyId,entityId)

              #if waterMoleculeInfo:
              #  for compIndexId in waterMoleculeInfo:
              #    self.sequences.append(NmrStarSequence(entitySaveFrame.title,self,entitySaveFrame,Id = entityAssemblyId, entityId = entityId, originalChainCode = originalChainCode, waterMoleculeCompIndexId = compIndexId, details = details, role = role, physState = physState, confIsomer = confIsomer, chemExchState = chemExchState, magnEquivCode = magnEquivCode))

              #else:
                self.sequences.append(NmrStarSequence(entitySaveFrame.title,self,entitySaveFrame,Id = entityAssemblyId, entityId = entityId, originalChainCode = originalChainCode, details = details, role = role, physState = physState, confIsomer = confIsomer, chemExchState = chemExchState, magnEquivCode = magnEquivCode))
                break

        #
        # Set molecular bonds
        #

        tableName = '_Bond'

        if tableName in self.saveFrame.tables:
          tableTags = self.saveFrame.tables[tableName].tags
          for i in range(len(tableTags['Type'])):

            bondType = tableTags['Type'][i]
            # TODO currently not handled!
            ###bondOrder = tableTags['Order'][i]

            sequenceElementData = []

            for j in range(1,3):
              for sequence in self.sequences:
                if sequence.Id == tableTags['Entity_assembly_ID_%d' % j][i]:
                  for sequenceElement in sequence.elements:
                    if sequenceElement.seqCode == tableTags['Comp_index_ID_%d' % j][i]:
                      sequenceElementData.append((sequenceElement,tableTags['Atom_ID_%d' % j][i]))

                  break

            if len(sequenceElementData) == 2:
              sequenceElementData[0][0].setBond(bondType,sequenceElementData[0][1],sequenceElementData[1][0],sequenceElementData[1][1])
              sequenceElementData[1][0].setBond(bondType,sequenceElementData[1][1],sequenceElementData[0][0],sequenceElementData[0][1])
            else:
              print ("  Warning: could not set bond information from NMR-STAR file, information missing.")


        tableName = '_Assembly_db_link'

        tagNames = ('Database_code','Accession_code','Entry_experimental_method','Entry_structure_resolution')

        if tableName in self.saveFrame.tables:
          tableTags = self.saveFrame.tables[tableName].tags

          for i in range(0,len(tableTags['Database_code'])):

            infoList = []
            for tagName in tagNames:
              infoList.append(tableTags[tagName][i])

            self.dbLinks.append(MolSystemDbLink(self,*infoList))

        tableName = '_Assembly_bio_function'

        tagNames = ('Biological_function',)

        if tableName in self.saveFrame.tables:
          tableTags = self.saveFrame.tables[tableName].tags

          for i in range(0,len(tableTags['Biological_function'])):

            infoList = []
            for tagName in tagNames:
              infoList.append(tableTags[tagName][i])

            self.bioFunctions.append(MolSystemBioFunction(self,*infoList))

    elif self.version == '2.1.1':

      # Handled completely differently from above! Also possible that molecular_system sf not present, will deal with this.

      moleculeName = self.saveFrame.title

      if 'molecular_system' in self.parent.sfs:
        molSysSaveFrame = self.parent.sfs['molecular_system'][0]
        tableTags = molSysSaveFrame.tables['_Mol_system_component_name'].tags

        for i in range(len(tableTags['_Mol_system_component_name'])):

          chainName = tableTags['_Mol_system_component_name'][i]
          moleculeLabel = tableTags['_Mol_label'][i]

          if moleculeLabel[1:] == moleculeName:
            self.sequences.append(NmrStarSequence(self.name,self,self.saveFrame, chainName = chainName))

      else:
        self.sequences.append(NmrStarSequence(self.name,self,self.saveFrame, chainName = self.defaultMolCode))

  def getWaterMoleculeInfo(self, entityAssemblyId, entityId):

    #
    # Special treatment for water molecules
    #

    waterMoleculeInfo = []

    tableName = '_PDBX_nonpoly_scheme'
    if self.version in ('3.1',) and tableName in self.saveFrame.tables:

      tableTags = self.saveFrame.tables[tableName].tags

      for i in range(len(tableTags['Entity_ID'])):

        if entityAssemblyId == tableTags['Entity_assembly_ID'][i] and entityId == tableTags['Entity_ID'][i]:

          if tableTags['Comp_ID'][i] == 'HOH':
            waterMoleculeInfo.append(tableTags['Comp_index_ID'][i])

    return waterMoleculeInfo

class NmrStarSequence:

  def __init__(self,name,parent,sequenceSaveFrame,Id = None,entityId = None, originalChainCode = None, waterMoleculeCompIndexId = None, details = None, role = None, physState = None, confIsomer = None, chemExchState = None, magnEquivCode = None, chainName = None):

    self.elements = []
    self.dbLinks = []
    self.bioFunctions = []

    self.isCircular = None

    self.saveFrame = sequenceSaveFrame

    self.parent = parent
    self.version = self.parent.version

    self.Id = Id
    self.entityId = entityId

    # 2.1.1 only
    self.chainName = chainName

    self.originalChainCode = originalChainCode
    self.waterMoleculeCompIndexId = waterMoleculeCompIndexId
    self.details = details
    self.role = role

    #print 'DATA: [%s] [%s] [%s]' % (physState, confIsomer, chemExchState)
    self.physState = physState
    self.confIsomer = confIsomer
    self.chemExchState = chemExchState

    self.magnEquivCode = magnEquivCode

    tagNames = {}
    tableTagNames = {}

    if self.version == '2.1.1':
      self.molName = self.formatMoleculeName(sequenceSaveFrame.title)

      tagNames['molType'] = '_Mol_type'
      tagNames['polymerType'] = '_Mol_polymer_class'
      tagNames['commonName'] = '_Name_common'
      tagNames['otherNames'] = '_Name_variant'
      tagNames['abbrev'] = '_Abbreviation_common'
      tagNames['molMass'] = '_Molecular_mass'

      tableName = '_Residue_seq_code'

      seqCodeTag = seqIdTag = '_Residue_seq_code'
      authorSeqCodeTag = '_Residue_author_seq_code'

      if ( tableName in self.saveFrame.tables and authorSeqCodeTag in
           self.saveFrame.tables[tableName].tags):

        validCode = 1
        sequenceTableTags = self.saveFrame.tables[tableName].tags
        for i in range(len(sequenceTableTags[authorSeqCodeTag])):

          if sequenceTableTags[authorSeqCodeTag][i] == None:
            validCode = 0
            break

        if validCode:
          seqCodeTag = authorSeqCodeTag

      tableTagNames['seqId'] = seqIdTag # This one to track the sequential numbering used in chem shift section
      tableTagNames['seqCode'] = seqCodeTag
      tableTagNames['code3Letter'] = '_Residue_label'

    elif self.version == '3.0':
      self.molName = self.formatMoleculeName(name)

      tagNames['molType'] = 'Type'
      tagNames['polymerType'] = 'Pol_type'
      #tagNames['commonName'] = '_Name_common'
      #tagNames['otherNames'] = '_Name_variant'
      #tagNames['abbrev'] = '_Abbreviation_common'
      #tagNames['molMass'] = '_Molecular_mass'

      """
      From RCSB:

      Data items in the ENTITY_POLY_SEQ category specify the sequence
      of monomers in a polymer. Allowance is made for the possibility
      of microheterogeneity in a sample by allowing a given sequence
      number to be correlated with more than one monomer id - the
      corresponding ATOM_SITE entries should reflect this
      heterogeneity.
      """

      # TODO: just use entity_poly_seq for now - should be OK...
      tableName = '_Entity_poly_seq'
      tableTagNames['seqCode'] = 'Comp_index_num'
      tableTagNames['code3Letter'] = 'Mon_ID'

    elif self.version == '3.1':

      if sequenceSaveFrame.tags['Name'] and sequenceSaveFrame.tags['Name'] != '.':
        self.molName = self.formatMoleculeName(sequenceSaveFrame.tags['Name'])
      else:
        self.molName = self.formatMoleculeName(name)

      tagNames['molType'] = 'Type'
      tagNames['polymerType'] = 'Polymer_type'
      tagNames['Mutation'] = 'Mutation'
      tagNames['EC_number'] = 'EC_number'
      tagNames['Fragment'] = 'Fragment'
      tagNames['Src_method'] = 'Src_method'
      #tagNames['commonName'] = '_Name_common'
      #tagNames['otherNames'] = '_Name_variant'
      #tagNames['abbrev'] = '_Abbreviation_common'
      #tagNames['molMass'] = '_Molecular_mass'

      tagNames['details'] = 'Details'
      tagNames['seqDetails'] = 'Polymer_author_seq_details'
      tagNames['authDefSeq'] = 'Polymer_author_defined_seq'

      tagNames['ambConfStates'] = 'Ambiguous_conformational_states'
      tagNames['ambChemCompSites'] = 'Ambiguous_chem_comp_sites'

      if sequenceSaveFrame.tags['Type'] in ('non-polymer','water'):
        tableName = '_Entity_comp_index'
        tableTagNames['seqCode'] = 'ID'
        tableTagNames['code3Letter'] = 'Comp_ID'
      else:
        tableName = '_Entity_comp_index'
        tableTagNames['seqCode'] = 'ID'
        tableTagNames['code3Letter'] = 'Comp_ID'
        tableTagNames['authorSeqCode'] = 'Auth_seq_ID'

        # Eldon thinks we can use the _Entity_comp_index table instead
        # and then also get the author residue numbers.

        #tableName = '_Entity_poly_seq'
        #tableTagNames['seqCode'] = 'Comp_index_ID'
        #tableTagNames['code3Letter'] = 'Mon_ID'

    #
    # Set tag names
    #

    for attrName in tagNames.keys():
      tagName = tagNames[attrName]
      if tagName in sequenceSaveFrame.tags:
        if sequenceSaveFrame.tags[tagName] != None:
          setattr(self,attrName,sequenceSaveFrame.tags[tagName])
        else:
          setattr(self,attrName,None)

    #
    # Set table tags (sequence elements)
    #

    resettingSeqCodes = False

    if tableName in self.saveFrame.tables:

      sequenceTableTags = self.saveFrame.tables[tableName].tags
      curSeqInsertCode = None

      for i in range(len(sequenceTableTags[tableTagNames['seqCode']])):

        if self.version == '3.1' and sequenceSaveFrame.tags['Type'] == 'water' and waterMoleculeCompIndexId:

          # TODO: HERE SHOULD BE FINDING CODE FROM COORDINATE SECTION! Use self.parent.origSeqCodeBySeqId
          if sequenceTableTags[tableTagNames['seqCode']][i] != waterMoleculeCompIndexId:
            continue

        # This is an on-the-fly conversion...
        bmrbCode = sequenceTableTags[tableTagNames['code3Letter']][i]
        ccpMapping = None

        if bmrbCode in bmrbCodeToCcpCode.keys():
          ccpMapping = bmrbCodeToCcpCode[bmrbCode]

          if ccpMapping[0]:
            self.elements.append(NmrStarSequenceElement(sequenceTableTags,i,tableTagNames))
            self.elements[-1].code3Letter = ccpMapping[0][1]
            self.elements[-1].residueType = ccpMapping[0][0] # This is usually not there!

        self.elements.append(NmrStarSequenceElement(sequenceTableTags,i,tableTagNames))

        #
        # Reset seqCode if can get it from coordinate section
        #

        if self.version == '3.1':

          residueKey = (self.Id,sequenceTableTags[tableTagNames['seqCode']][i])

          #
          # Now set the sequence code...
          #

          if self.parent.origSeqCodeBySeqId:
            #sks = self.parent.origSeqCodeBySeqId.keys()
            #sks.sort()
            #  print sk, self.parent.origSeqCodeBySeqId[sk]
            #for sk in sks:
            #sys.exit()
            if residueKey in self.parent.origSeqCodeBySeqId:

              #
              # Set sequence insertion code specifically for this residue
              #

              if self.parent.origSeqInsertCodeBySeqId and residueKey in self.parent.origSeqInsertCodeBySeqId:
                self.elements[-1].insertionCode = self.parent.origSeqInsertCodeBySeqId[residueKey]
                curSeqInsertCode = self.parent.origSeqInsertCodeBySeqId[residueKey]
              else:
                curSeqInsertCode = None


              resettingSeqCodes = True
              self.elements[-1].seqCode = self.parent.origSeqCodeBySeqId[residueKey]
              #print "RESET ORIG for %s to %d" % (residueKey,self.elements[-1].seqCode)

              # Also try and reset previous sequence codes based on coordinate section
              # This is to prevent overlapping sequence codes, and makes code work better in general.
              # Only do this if no sequence insertion code was used in the previous residue...
              for i in range(1,99999):
                if len(self.elements) <= i:
                  break
                prevResidueKey = (residueKey[0],residueKey[1] - i)
                if not prevResidueKey in self.parent.origSeqCodeBySeqId:
                  self.elements[-(i+1)].seqCode = self.elements[-1].seqCode - i
                  #print "RESET RANGE", i, residueKey[1]-i, self.elements[-(i+1)].seqCode
                else:
                  break

            elif resettingSeqCodes:
              prevSeqCode = self.elements[-2].seqCode
              self.elements[-1].seqCode = prevSeqCode + 1
              #print "RESET PREV", (prevSeqCode + 1)

            if curSeqInsertCode:
              self.elements[-1].insertionCode = curSeqInsertCode

          #
          # Also set the atom names
          #

          if self.parent.atomNamesBySeqId and residueKey in self.parent.atomNamesBySeqId:
            self.elements[-1].setAtomNames(self.parent.atomNamesBySeqId[residueKey])

        # Set mapping if required.
        if ccpMapping:
          self.elements[-1].code3Letter = ccpMapping[1][1]
          self.elements[-1].residueType = ccpMapping[1][0]

          if ccpMapping[2]:
            self.elements.append(NmrStarSequenceElement(sequenceTableTags,i,tableTagNames))
            self.elements[-1].code3Letter = ccpMapping[2][1]
            self.elements[-1].residueType = ccpMapping[2][0] # This is usually not there!

    """
    print "SEQUENCE"
    for seqEl in self.elements:
      print seqEl.seqCode,
      if hasattr(seqEl, 'insertionCode'):
        print seqEl.insertionCode
      else:
        print
    """

    #
    # Get molecule database links.
    #

    tableName = '_Entity_db_link'
    tagNames = ('Database_code','Accession_code','Entry_mol_name','Entry_details')

    if tableName in self.saveFrame.tables:
      tableTags = self.saveFrame.tables[tableName].tags

      for i in range(0,len(tableTags['Database_code'])):

        infoList = []
        for tagName in tagNames:
          infoList.append(tableTags[tagName][i])

        self.dbLinks.append(MoleculeDbLink(self,*infoList))

    #
    # Get molecule database links.
    #

    tableName = '_Entity_biological_function'
    tagNames = ('Biological_function',)

    if tableName in self.saveFrame.tables:
      tableTags = self.saveFrame.tables[tableName].tags

      for i in range(0,len(tableTags['Biological_function'])):

        infoList = []
        for tagName in tagNames:
          infoList.append(tableTags[tagName][i])

        self.bioFunctions.append(MoleculeBioFunction(self,*infoList))

  def formatMoleculeName(self,molName):

    # Have to make sure that this is not too long, only 255 chars allowed in CCPN. Truncating to
    # 250 so some extra place for adding chars in sequence read pipeline

    return molName[:65]

class NmrStarSequenceElement:
  """
  Information for an element in the sequence (residue)
  """
  def __init__(self,sequenceTableTags,i,tableTagNames):

    self.bonds = {}
    self.formatCode = None

    self.atomNames = []

    self.secStrucInfo = {}

    for attrName in tableTagNames.keys():
      if tableTagNames[attrName] in sequenceTableTags and len(sequenceTableTags[tableTagNames[attrName]]) > i and sequenceTableTags[tableTagNames[attrName]][i] is not None:
        setattr(self,attrName,sequenceTableTags[tableTagNames[attrName]][i])
      else:
        setattr(self,attrName,None)

  def setAtomNames(self,atomNames):

    self.atomNames = atomNames

  def setBond(self,bondType,atomName,bondedSeqEl,bondedAtomName):

    if not bondType in self.bonds:
      self.bonds[bondType] = {}

    if not atomName in self.bonds[bondType]:
      self.bonds[bondType][atomName] = []

    if not (bondedSeqEl,bondedAtomName) in self.bonds[bondType][atomName]:
      self.bonds[bondType][atomName].append((bondedSeqEl,bondedAtomName))
      print ("  Found %s bond from %s.%s - %s.%s" % (bondType,self.seqCode,atomName,bondedSeqEl.seqCode,bondedAtomName))

# TODO - move these to projectIO or generalIO - it complained earlier due to circular references I think?

class MoleculeDbLink:

  def __init__(self,parent,dbName,dbAcc,dbMolName,dbDetails):

    self.parent = parent
    self.dbName = dbName
    self.dbAcc = dbAcc
    self.dbMolName = dbMolName
    self.dbDetails = dbDetails

class MolSystemDbLink:

  def __init__(self,parent,dbName,dbAcc,exptMethod,structResolution):

    self.parent = parent
    self.dbName = dbName
    self.dbAcc = dbAcc
    self.exptMethod = exptMethod
    self.structResolution = structResolution

class MoleculeBioFunction:

  def __init__(self,parent,bioFunc):

    self.parent  = parent
    self.bioFunc = bioFunc

class MolSystemBioFunction:

  def __init__(self,parent,bioFunc):

    self.parent  = parent
    self.bioFunc = bioFunc



import glob, os, json
import numpy as np
from dynaMine.analysis.parsers import DynaMineRelatedFileParsers

class mapPredToMSA(DynaMineRelatedFileParsers):

  predictionTypes = ('backbone','coil','helix','ppII','sheet','sidechain','earlyFoldProb')

  gapCode = "-"

  def getPredictionFileName(self, seqId, predictionType):

    # This needs to be redefined for other predictions in a subclass!

    subDir = "dynaMineResults/{}/{}_uniref90.clu/{}_uniref90.clu_{}.pred"
    #subDir = "dynaMineResults/{}/{}_uniref90_msa/{}_uniref90_msa_{}.pred"

    predictionFileName = subDir.format(seqId,seqId,seqId,predictionType)

    return predictionFileName

  def doMapping(self,msaDir):

    alignFiles = glob.glob("{}/*.clustal".format(msaDir))

    for alignFile in alignFiles:

      uniprotId = os.path.split(alignFile)[1].replace("_uniref90.clustal","")
      #uniprotId = os.path.split(alignFile)[1].replace("_uniref90_msa.txt","")

      if uniprotId != 'P0DTC9':
        continue

      # Get alignment info
      # Is dictionary with {seqId: alignment} entries
      seqAlignments = self.readAlignments(alignFile,resetAlignRefSeqID=True)

      # Get predictions
      predictions = {}
      print("Working on {}".format(uniprotId))
      for predictionType in self.predictionTypes:

        predFile = self.getPredictionFileName(uniprotId,predictionType)

        assert os.path.exists(predFile), "File missing {}".format(predFile)

        predictions[predictionType] = self.readDynaMineMultiEntry(predFile)

      # Map them
      alignedPredictions = {}
      for predictionType in self.predictionTypes:
        alignedPredictions[predictionType] = {}

      allSeqIds = list(seqAlignments.keys())
      allSeqIds.sort()

      if uniprotId not in allSeqIds:
        print('Key ID missing for {}!'.format(uniprotId))
        continue

      sequenceInfo = {}

      for seqId in allSeqIds:

        alignment = seqAlignments[seqId]
        seqIndex = 0

        sequenceInfo[seqId] = []

        for predictionType in self.predictionTypes:
          alignedPredictions[predictionType][seqId] = []

        for alignIndex in range(len(alignment)):
          if alignment[alignIndex] == self.gapCode:
            for predictionType in self.predictionTypes:
              alignedPredictions[predictionType][seqId].append(None)
          else:
            resName = predictions['backbone'][seqId][seqIndex][0]

            assert resName == alignment[alignIndex] or resName == 'X', "Amino acid code mismatch {}-{}".format(resName,alignment[alignIndex])

            sequenceInfo[seqId].append(alignment[alignIndex])

            for predictionType in self.predictionTypes:
              (resName,predValue) = predictions[predictionType][seqId][seqIndex]
              alignedPredictions[predictionType][seqId].append(predValue)
            seqIndex += 1

      alignedPredictions['sequence'] = sequenceInfo

      distribKeys = ('median','thirdQuartile','firstQuartile','topOutlier','bottomOutlier')
      numDistribKeys = len(distribKeys)

      # Now generate the info for quartiles, ... based on the alignRefSeqID, first entry in alignment file
      refSeqPredictionDistribs = {}
      for predictionType in self.predictionTypes:
        refSeqPredictionDistribs[predictionType] = {self.alignRefSeqID: []}
        for distribKey in distribKeys:
          refSeqPredictionDistribs[predictionType][distribKey] = []

      self.alignRefSeqID = uniprotId
      refSeqAlignment = seqAlignments[self.alignRefSeqID]
      refSeqIndex = 0
      for alignIndex in range(len(refSeqAlignment)):
        if refSeqAlignment[alignIndex] == self.gapCode:
          continue
        else:
          for predictionType in self.predictionTypes:
            predValues = [alignedPredictions[predictionType][seqId][alignIndex] for seqId in allSeqIds if alignedPredictions[predictionType][seqId][alignIndex] != None]

            distribInfo = self.getDistribInfo(predValues)

            refSeqPredictionDistribs[predictionType][self.alignRefSeqID].append(alignedPredictions[predictionType][self.alignRefSeqID][alignIndex])

            for i in range(numDistribKeys):
              refSeqPredictionDistribs[predictionType][distribKeys[i]].append(distribInfo[i])

          refSeqPredictionDistribs['sequence'] = sequenceInfo[self.alignRefSeqID]

          refSeqIndex += 1

      # Ready, dump to JSON
      self.dumpJsonToFile(refSeqPredictionDistribs,"Disomine_byAlignment/{}_forRefSeq.json".format(self.alignRefSeqID))
      self.dumpJsonToFile(alignedPredictions,"Disomine_byAlignment/{}_fullAlignment.json".format(self.alignRefSeqID))

  def dumpJsonToFile(self,dictToDump, outFile):

      with open(outFile, 'w+') as outfile:
        json.dump(dictToDump, outfile, indent=4)

  def getDistribInfo(self,valueList, outlierConstant = 1.5):

    median = np.median(valueList)
    upper_quartile = np.percentile(valueList, 75)
    lower_quartile = np.percentile(valueList, 25)
    IQR = (upper_quartile - lower_quartile) * outlierConstant

    return (median, upper_quartile, lower_quartile,upper_quartile + IQR,lower_quartile - IQR)

if __name__ == '__main__':

  mps = mapPredToMSA()

  mps.doMapping("sars-cov-2/msa_200408")
  #mps.doMapping("sars-cov-2-msa_files")





import json
import numpy as np
from sklearn import mixture

class MsaGmmAssessment():

    def gmm_generator(self, msa_filename, single_preds):
        data_full = []
        with open(msa_filename) as json_file:
            data = json.load(json_file)
            for aln_pos in range(len(data['results']['results']['backbone']['median'])):
                pred_vector = []
                for biophys in ['backbone', 'sidechain', 'coil', 'sheet', 'helix',
                     'earlyFolding', 'disoMine']:
                     pred_vector.append(
                        data['results']['results'][biophys]['median'][aln_pos])

                data_full.append(pred_vector)

        X_train = np.vstack(data_full)
        clf = mixture.GaussianMixture(n_components = 1, covariance_type = 'full',
                                      verbose = 2, verbose_interval = 1)
        clf.fit(X_train)

        gmm_dict = {}
        scores = []
        gmm_info = {}
        for prot in single_preds.keys():
            if prot != "sequence":
                full_pred = []
                for res in range(len(single_preds[prot]['backbone'])):
                    pred_vector = []
                    for biophys in ['backbone', 'sidechain', 'coil', 'sheet', 'helix',
                                    'earlyFolding', "disoMine"]:
                        if single_preds[prot][biophys][res] != None:
                            pred_vector.append(single_preds[prot][biophys][res])

                    full_pred.append(pred_vector)

                full_pred = [list for list in full_pred if list]

                preds = np.vstack(full_pred)
                gmm_info[prot] = clf.score_samples(preds).tolist()
                scores.extend(clf.score_samples(preds).tolist())

        perc_95 = np.percentile(scores, 5)
        perc_99 = np.percentile(scores, 1)
        perc_99_9 = np.percentile(scores, 0.1)

        for key in gmm_info.keys():
            perc_95_list = ["Out" if i <= perc_95 else "In" for i in gmm_info[key]]
            perc_99_list = ["Out" if i <= perc_99 else "In" for i in gmm_info[key]]
            perc_99_9_list = ["Out" if i <= perc_99_9 else "In" for i in gmm_info[key]]

            gmm_dict[key] = {"scores": gmm_info[key], "perc_95": perc_95_list,
                   "perc_99": perc_95_list, "perc_99_9": perc_99_9_list}

        with open('{}_quantification.json'.format(msa_filename.split(".")[0]), 'w') as fp:
            json.dump(gmm_dict, fp)



#!/usr/bin/env python
# encoding: utf-8
"""
Predictor.py

Created by Wim Vranken on 1st May 2020, as wrapper around original PSPer code
Note that the original DisoMine code was upgraded to Python3.7 and simplified by taking out redundant code for production version
The DisoMine code was also modified to better work in conjunction with the data structures used here.
Copyright (c) 2020- Wim Vranken. All rights reserved.
"""

from b2bTools.general.Io import B2bIo
from b2bTools.singleSeq.PSPer.standalone import standalone
import numpy as np

class PSPer(B2bIo):

  version = '1.0'
  scriptName = "b2bTools.singleSeq.PSPer.Predictor"
  name ="PSPer"

  def __init__(self):
    # Additional info for writing files
    self.references = ['Not yet available, submitted']
    self.infoTexts = ['Generated by PSPer', 'See http://bio2byte.be']
    self.allPredictions = {}
    self.informationPerPredictor = {
      self.name: {
        'references': self.references,
        'info':       ";".join(self.infoTexts),
        'version':    self.version,
        'origin':     self.scriptName
      }
    }

  # Warning: THIS IS NOW NOT WORKING FROM A DICT BUT FROM A LIST OF TUPLES!
  def predictSeqs(self, seqs):
      psper_results = standalone(seqs)

      if not psper_results:
          print("PSPer failed....")
          return
      elif 'error' in psper_results:
          print("PSPer failed with known error....", psper_results['error'])
          return

      psp_results = {}
      sequences_count = len(psper_results['results'])
      for seq_index in range(sequences_count):
            seq_dict = psper_results['results'][seq_index]
            key = seq_dict['proteinID']
            del seq_dict['proteinID']

            sequence = [residue for residue in seq_dict['sequence']]
            del seq_dict['sequence']

            psp_results[key] = { 'seq': sequence }

            for prediction_key, prediction_value in seq_dict.items():
                if type(prediction_value) == list or type(prediction_value) == np.array or type(prediction_value) == np.ndarray:
                    psp_results[key][prediction_key] = [(residue, prediction) for (residue, prediction) in zip(sequence, prediction_value)]
                else:
                    psp_results[key][prediction_key] = prediction_value

      self.allPredictions = psp_results



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

prop_sw = 35

import os
from .runpsipred_single import run

### objects ###
amino_index = {'A': 1, 'C': 2, 'B': -1, 'E': 3, 'D': 4, 'G': 5, 'F': 6, 'I': 7, 'H': 8, 'K': 9, 'M': 10, 'L': 11,
               'N': 12, 'Q': 13, 'P': 14, 'S': 15, 'R': 16, 'U': -1, 'T': 17, 'W': 18, 'V': 0, 'Y': 19, 'X': -1,
               'Z': -1}


class psipred:

    def __init__(self, tmpfold, single_sequence=True):

        self.tmpfold = tmpfold

        if single_sequence:
            pass
        else:
            raise (NotImplementedError)

    def predict(self, seq, seq_ID, base_name):
        pred = []

        fasta_filename = os.path.join(self.tmpfold, base_name + '.fasta')
        ss2_filename = os.path.join(self.tmpfold, base_name + '.ss2')
        horiz_filename = os.path.join(self.tmpfold, base_name + '.horiz')
        ss_filename = os.path.join(self.tmpfold, base_name + '.ss')
        mtx_filename = os.path.join(self.tmpfold, base_name + '.mtx')

        with open(fasta_filename, 'w') as fasta_file:
            fasta_file.write('>{0}\n{1}'.format(seq_ID, seq))

        # Executes PSIPRED binary
        run(fasta_filename, self.tmpfold)

        # Reads the PSIPRED output file .ss2
        with open(ss2_filename, 'r') as ss2_file:
            ss2 = ss2_file.readlines()

        for i in ss2[2:]:
            a = i.strip().split()

            if len(a) > 0:
                pred += [[float(a[3]), float(a[4]), float(a[5])]]

        # Deletes the temp files
        if os.path.exists(fasta_filename):
            os.remove(fasta_filename)

        if os.path.exists(ss2_filename):
            os.remove(ss2_filename)

        if os.path.exists(horiz_filename):
            os.remove(horiz_filename)

        if os.path.exists(ss_filename):
            os.remove(ss_filename)

        if os.path.exists(mtx_filename):
            os.remove(mtx_filename)

        return pred


def build_vector(seq, dmPredictions, base_name=None, seqID=None, TYPE=None, sw=None, nomeseq=None, tmpfold=None):
    vector = []
    sequence_length = len(seq)
    last = 0
    seq_nogap = seq.replace('-', '')

    for i in range(sequence_length):
        vector += [[]]

    nfeatures = 0

    for curr_fea in TYPE.split(','):
        if curr_fea.startswith('dyna') or curr_fea == 'ef':

            if curr_fea == 'dyna_coil':
                v_dyna = [dmPredictions['coil']]
            elif curr_fea == 'dyna_sheet':
                v_dyna = [dmPredictions['sheet']]
            elif curr_fea == 'dyna_helix':
                v_dyna = [dmPredictions['helix']]
            elif curr_fea == 'dyna_side':
                v_dyna = [dmPredictions['sidechain']]
            elif curr_fea == 'dyna_back':
                v_dyna = [dmPredictions['backbone']]
            elif curr_fea == 'ef':
                v_dyna = [dmPredictions['earlyFolding']]

            effect = 0  ## We could remove it and only use vector_index. Validate hypothesis.

            for vector_index in range(len(vector)):
                if seq[vector_index] != '-':
                    # print i
                    for s in range(-sw, sw + 1):
                        if effect + s < 0:
                            vector[vector_index] += [0]
                        elif effect + s >= len(seq_nogap):
                            vector[vector_index] += [0]
                        else:
                            vector[vector_index] += [v_dyna[0][effect + s][1]]

                    effect += 1
                    last = i

        elif 'psipred' == curr_fea:
            if tmpfold == None:
                # Directory of this script
                scriptDir = os.path.dirname(os.path.realpath(__file__))
                tmpfold = os.path.join(scriptDir, 'psipred/tmp/')

            psidpred_runner = psipred(tmpfold)
            v_psipred = psidpred_runner.predict(seq_nogap, seqID, base_name=base_name)
            effect = 0

            for vector_index in range(len(vector)):
                if seq[vector_index] != '-':
                    for s in range(-sw, sw + 1):
                        if effect + s < 0:
                            vector[vector_index] += [0] * 3
                        elif effect + s >= len(seq_nogap):
                            vector[vector_index] += [0] * 3
                        else:
                            vector[vector_index] += v_psipred[effect + s]

                    effect += 1
                    last = i

        else:
            print((curr_fea, 'IS NOT A FEATURE'))
            assert False

    for sequence_index in range(sequence_length):
        if seq[sequence_index] == '-' or seq[sequence_index] == '.':
            vector[sequence_index] += ['-'] * nfeatures

    return vector



#!/usr/bin/python

"""
======================COPYRIGHT/LICENSE START==========================

rdcConstraintsIO.py: I/O for nmrStar rdc constraints saveframes

Copyright (C) 2005 Wim Vranken (European Bioinformatics Institute)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

A copy of this license can be found in ../../../../license/LGPL.license

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

- contact Wim Vranken (wim@ebi.ac.uk)
=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

===========================REFERENCE END===============================
"""

import os

from b2bTools.general.ccpn.universal.Io import getTopDirectory

from b2bTools.general.ccpn.format.nmrStar.generalIO import NmrStarConstraintFile
from b2bTools.general.ccpn.format.nmrStar.generalIO import NmrStarFile
from b2bTools.general.ccpn.format.nmrStar.generalIO import GenericConstraint

#
# For DIANA torsion angle name information
#

#from b2bTools.general.ccpn.format.cyana.cyanaLibParser import CyanaLibrary

from b2bTools.general.ccpn.format.general.Constants import defaultMolCode
from b2bTools.general.ccpn.format.general.Util import getSeqAndInsertCode

#####################
# Class definitions #
#####################


class NmrStarFile(NmrStarFile):

  def initialize(self, version = '2.1.1'):

    self.constraintFiles = []
    self.files = self.constraintFiles

    if not self.version:
      self.version = version

    if self.version == '3.1':
      self.saveFrameName = 'RDC_constraints'
    else:
      self.saveFrameName = 'residual_dipolar_couplings'

    self.DataClassFile = NmrStarRdcConstraintFile

    self.components = [self.saveFrameName]
    self.setComponents()

  def read(self,verbose = 0):

    self.readComponent(verbose = verbose)

class NmrStarRdcConstraintFile(NmrStarConstraintFile):

  def initialize(self,parent,saveFrame = None):

    self.constraints = []

    self.constraintElements = 2

    self.cyanaLibUsed = 0
    self.cyanaLib = CyanaLibrary()

    self.saveFrame = saveFrame

    self.parent = parent
    self.version = parent.version

    if self.saveFrame:
      self.parseSaveFrame()

  def parseSaveFrame(self):

    if not self.checkVersion():
      return

    #
    # Set up reference info
    #

    tableTagNames = {}

    if self.version == '2.1.1':

      tableName = '_Residual_dipolar_coupling_ID'
      tableTagNames['ID'] = tableName

      tableTagNames['atomPresent'] = "_Atom_two_atom_name"
      tableTagNames['position'] = ['one','two']

      tableTagNames['code'] = '_Residual_dipolar_coupling_code'
      tableTagNames['value'] = '_Residual_dipolar_coupling_value'

      #
      # Hack - both of these occur in the files for some reason
      #

      if self.saveFrame.tables.has_key(tableName):

        constraintTableTags = self.saveFrame.tables[tableName].tags

        # This value not necessarily available...
        if constraintTableTags.has_key('_Residual_dipolar_coupling_value_error'):
          tableTagNames['error'] = '_Residual_dipolar_coupling_value_error'
        elif constraintTableTags.has_key('_Residual_dipolar_coupling_error'):
          tableTagNames['error'] = '_Residual_dipolar_coupling_error'

      tableTagNames['lowerValue'] = '_Residual_dipolar_coupling_lower_bound_value'
      tableTagNames['upperValue'] = '_Residual_dipolar_coupling_upper_bound_value'

      # Same as in dihedralConstraintsIO!
      tableTagNames['chainCode'] = "_Atom_%s_mol_system_component_name"
      tableTagNames['seqCode'] = "_Atom_%s_residue_seq_code"
      tableTagNames['atomName'] = "_Atom_%s_atom_name"
      tableTagNames['resLabel'] = "_Atom_%s_residue_label"

    elif self.version[0] == '3':

      tableName = '_RDC_constraint'
      tableTagNames['ID'] = 'ID'

      tableTagNames['position'] = ['1','2']

      #tableTagNames['code'] = '_Residual_dipolar_coupling_code'
      tableTagNames['value'] = 'RDC_val'
      tableTagNames['error'] = 'RDC_val_err'
      tableTagNames['lowerValue'] = 'RDC_lower_bound'
      tableTagNames['upperValue'] = 'RDC_upper_bound'

      if self.version == '3.0':
        tableTagNames['atomPresent'] = "Label_atom_ID_2"
        tableTagNames['chainCode'] = "Label_entity_assembly_ID_%s"
        tableTagNames['seqCode'] = "Label_comp_index_ID_%s"
        tableTagNames['atomName'] = "Label_atom_ID_%s"
        tableTagNames['resLabel'] = "Label_comp_ID_%s"

      elif self.version == '3.1':
        tableTagNames['atomPresent'] = "Atom_ID_2"
        tableTagNames['chainCode'] =   "Entity_assembly_ID_%s"
        tableTagNames['seqCode'] =     "Comp_index_ID_%s"
        tableTagNames['atomName'] =    "Atom_ID_%s"
        tableTagNames['resLabel'] =    "Comp_ID_%s"

      if self.saveFrame.tables.has_key(tableName) and not self.saveFrame.tables[tableName].tags[tableTagNames['seqCode'] % 1][0]:
        tableTagNames['atomPresent'] = "Auth_atom_ID_2"

        if self.version == '3.0':
          tableTagNames['chainCode'] = "Auth_segment_code_%s"
        elif self.version == '3.1':
          tableTagNames['chainCode'] = "Auth_asym_ID_%s"

        tableTagNames['seqCode'] = "Auth_seq_ID_%s"
        tableTagNames['atomName'] = "Auth_atom_ID_%s"
        tableTagNames['resLabel'] = "Auth_comp_ID_%s"

    #
    # Could be missing...
    #

    if self.saveFrame.tables.has_key(tableName):

      constraintTableTags = self.saveFrame.tables[tableName].tags

      for i in range(0,len(constraintTableTags[tableTagNames['ID']])):

        rdcConstraint = NmrStarRdcConstraint()
        rdcConstraint.setData(constraintTableTags,i,tableTagNames)

        self.constraints.append(rdcConstraint)

        rdcConstraint.setErrors(self.saveFrame.tables[tableName].tagErrors,tableTagNames,i)

        if rdcConstraint.Id and not constraintTableTags[tableTagNames['atomPresent']][i]:

          #
          # Only one atom given - find other one... also set code to something for
          # output later on (hack for writing constraint nmrStar files for Jurgen)
          #

          rdcConstraint.setAtomMembers(constraintTableTags,i,self.cyanaLib,tableTagNames)
          rdcConstraint.code = True

          self.cyanaLibUsed = 1

        else:

          rdcConstraint.nodes.append(NmrStarRdcConstraintItem())

          for position in tableTagNames['position']:

            rdcConstraint.nodes[-1].members.append(NmrStarRdcConstraintMember())
            rdcConstraint.nodes[-1].members[-1].addStarInfo(constraintTableTags,i,position,tableTagNames)

    #
    # Comments
    #

    self.parseCommentsLoop()

class NmrStarRdcConstraint(GenericConstraint):

  def __init__(self,Id = None):

    self.Id = Id
    self.nodes = []   # THIS IS A HACK! Not really a node...
    self.errors = []
    self.code = None

  def setData(self,constraintLogicTableTags,i,tableTagNames):

    self.Id = constraintLogicTableTags[tableTagNames['ID']][i]

    self.value = constraintLogicTableTags[tableTagNames['value']][i]

    if tableTagNames.has_key('error'):
      self.error = constraintLogicTableTags[tableTagNames['error']][i]
    else:
      self.error = None

    if constraintLogicTableTags.has_key(tableTagNames['lowerValue']):
      self.lowerValue = constraintLogicTableTags[tableTagNames['lowerValue']][i]

    if constraintLogicTableTags.has_key(tableTagNames['upperValue']):
      self.upperValue = constraintLogicTableTags[tableTagNames['upperValue']][i]

    if tableTagNames.has_key('code') and constraintLogicTableTags.has_key(tableTagNames['code']):
      self.code = constraintLogicTableTags[tableTagNames['code']][i]

  def setAtomMembers(self,constraintTableTags,i,cyanaLib,tableTagNames):

    #
    # Need to set two members... for DIANA-CYANA only one available.
    # Using DIANA library to find out where the 'other' atom is...
    #

    chainCode = constraintTableTags[tableTagNames['chainCode'] % tableTagNames['position'][0]][i]

    if not chainCode:
      chainCode = defaultMolCode

    seqCode = constraintTableTags[tableTagNames['seqCode'] % tableTagNames['position'][0]][i]
    resLabel = constraintTableTags[tableTagNames['resLabel'] % tableTagNames['position'][0]][i]
    refAtomName = constraintTableTags[tableTagNames['atomName'] % tableTagNames['position'][0]][i]

    refAtom = cyanaLib.findAtom(resLabel,refAtomName)

    if not refAtom:
      print ("  Error: atom %s (%s %s, chain '%s') not found in Cyana library!" % (refAtomName,resLabel,seqCode,chainCode))

    elif refAtom.bondedAtomSerials.count(0) != 3:
      print ("  Error: invalid single atom %s (%s %s, chain '%s'). No or multiple bonded atoms" % (refAtomName,resLabel,seqCode,chainCode))

    else:

      self.nodes.append(NmrStarRdcConstraintItem())

      self.nodes[-1].members.append(NmrStarRdcConstraintMember(chainCode,seqCode,refAtomName,resLabel))

      atomSerial = refAtom.bondedAtomSerials[0]
      bondedAtom = cyanaLib.findAtomBySerial(resLabel,atomSerial)

      self.nodes[-1].members.append(NmrStarRdcConstraintMember(chainCode,seqCode,bondedAtom.name,resLabel))

class NmrStarRdcConstraintItem:

  def __init__(self):

    self.Id = 1
    self.members = []

class NmrStarRdcConstraintMember:

  def __init__(self,chainCode = None,seqCode = None, atomName = None, resLabel = None):

    self.chainCode = chainCode
    (self.seqCode,self.seqInsertCode) = getSeqAndInsertCode(seqCode)
    self.atomName = atomName
    self.resLabel = resLabel

    self.defaultMolCode = defaultMolCode

  def addStarInfo(self,constraintTableTags,i,position,tableTagNames):

    # Tag can be missing in 2.1.1 files...
    if constraintTableTags.has_key(tableTagNames['chainCode'] % position):
      self.chainCode = constraintTableTags[tableTagNames['chainCode'] % position][i]
    else:
      self.chainCode = None

    if not self.chainCode:
      self.chainCode = self.defaultMolCode

    (self.seqCode,self.seqInsertCode) = getSeqAndInsertCode(constraintTableTags[tableTagNames['seqCode'] % position][i])
    self.atomName = constraintTableTags[tableTagNames['atomName'] % position][i]
    self.resLabel = constraintTableTags[tableTagNames['resLabel'] % position][i]

###################
# Main of program #
###################

if __name__ == "__main__":

  files = [#['../reference/ccpNmr/jurgenBmrb/1jwe/restraints.star','2.1.1'],
           #['../reference/ccpNmr/jurgenBmrb/1j7p/restraints.star','2.1.1']
           ['../reference/ccpNmr/jurgenBmrb/1b4c/1b4c.str','3.0']
          ]

  for (file,version) in files:

    file = os.path.join(getTopDirectory(), file)

    nmrStarFile = NmrStarFile(file,version = version)

    nmrStarFile.read(verbose = 1)

    for constraintFile in nmrStarFile.constraintFiles:
      for constraint in constraintFile.constraints:

        print (constraint.Id, constraint.value, constraint.error, constraint.lowerValue, constraint.upperValue)

        for item in constraint.nodes:
          mlist = []
          for member in item.members:
            mlist.append([member.seqCode,member.atomName])
          print ("   " + str(mlist))



"""
Classes for dealing with STAR syntax
"""

__author__    = "$Author: jurgenfd $"
___revision__ = "$Revision: 1.10.4.1 $"
___date__     = "$Date: 2011-04-06 09:00:38 $"

"""
$Log: not supported by cvs2svn $
Revision 1.10  2007/07/09 12:22:26  wimvranken
Merge with branch 4.

Revision 1.9.4.1.4.4  2007/06/22 10:21:43  wimvranken
Fixed small string formatting bug

Revision 1.9.4.1.4.3  2007/06/13 08:35:52  wimvranken
Removed copyright - is not CCPN

Revision 1.9.4.1.4.2  2006/10/31 16:04:44  wimvranken
Removed crap char

Revision 1.9.4.1.4.1  2006/10/31 12:44:41  wimvranken
Output now formatted

Revision 1.9.4.1  2006/07/21 17:29:01  wimvranken
Now only contain full path imports

Revision 1.9  2005/11/03 14:07:19  wimvranken
Added extra error message printout

Revision 1.8  2005/06/27 16:43:42  wb104
Updated licenses.

Revision 1.7  2004/12/15 17:57:20  tjs23
TJS: Updated licenses.

Revision 1.6  2003/10/03 11:42:56  wfv20
Removed loose print statement

Revision 1.5  2003/09/25 12:12:49  wfv20
Adapted for better handling of quotes

Revision 1.4  2003/08/08 12:46:28  wfv20
Esthetic changes only

Revision 1.3  2003/07/24 15:28:53  wfv20
Debug code for loop crash added

Revision 1.2  2003/07/14 09:34:32  wfv20
Modified so list references are not carried through class initialization

Revision 1.1  2003/07/01 12:56:19  wfv20
Jurgen Doreleijers modified Python nmrStar reader. Bug fixes and additions (notably for reading nmrView star files) have been made.

RRevision 1.1.1.1  2001/11/02 20:16:40  jurgen
IInitial package capable of read/write access to STAR files without nested loops

"""
## Standard modules
import re

from .Utils import Lister
from .Text import *

"""
Looped and free tags can not be mixed in same object.
"""
class TagTable (Lister):
    """
    In initializing the class a content has to be given!!!
    If not then the class will make something up and it won't
    be pretty but it will be following legal syntax. Usual
    case is to call the parse method with some text and position.
    """

    floatPatt = re.compile("^(\-?\d+)\.?(\d*)$")
    integerPatt = re.compile("^\d+$")

    colspacer = 3 * ' '

    def __init__( self,
                  free      = None,
                  title     = '',
                  tagnames  =  None,
                  tagvalues =  None,
                  verbosity = 2
                  ):
        self.free       = free
        self.title      = title

        # Modified tagnames, tagvalues initialization so list references
        # are not carried through (Wim 14/07/2002)

        self.tagnames = tagnames
        if self.tagnames == None:
          self.tagnames  = [ '_Dummy_tag' ]

        self.tagvalues = tagvalues
        if self.tagvalues == None:
          self.tagvalues  = [ [None] ]

        self.verbosity  = verbosity

    "Returns the STAR text representation"
    def star_text ( self,
                    flavor                  = 'NMR-STAR'
                   ):

        #
        # Code modified for better output formatting - Wim 2006/10/25
        #

        ## Info herein can be transferred to a STAR reference file too
        if flavor == None or flavor == 'NMR-STAR':
            # Number of spaces before the loop_ tag. 0 in CIF
            loop_ident_size     = 3
            free_ident_size     = loop_ident_size     #@UnusedVariable
            tagnames_ident_size = loop_ident_size + 3 #@UnusedVariable
            show_stop_tag       = 1 #@UnusedVariable
        elif flavor == 'mmCIF':
            loop_ident_size     = 0
            tagnames_ident_size = 0 #@UnusedVariable
            show_stop_tag       = None #@UnusedVariable
        else:
            print('ERROR: Unknown flavor of STAR given', flavor)
            return 1

        loop_ident_size         = 3
        free_ident_size         = loop_ident_size
        tagnames_ident_size     = loop_ident_size + 3
        show_stop_tag           = 1

        starString  = ''
        count       = -1
        count_hash  = 100000 # Show progress hashes while composing text for each count_hash number of values approximately

        ## Free tags here
        if self.free:
            i = 0

            # Add some formatting...
            maxTagName = 0
            for tagName in self.tagnames:
              if len(tagName) > maxTagName:
                maxTagName = len(tagName)

            tagNameFormat = "%-" + str(maxTagName + 2) + "s"

            for tagname in self.tagnames:
                ## Just format it such that it will take the least space
                if pattern_quotes_needed.search( self.tagvalues[i][0] ):
                    tagvalue = quotes_add( self.tagvalues[i][0] )
                else:
                    tagvalue = self.tagvalues[i][0]
                starString += free_ident_size * ' ' + (tagNameFormat % tagname) + " %s" % ( tagvalue )
                if tagvalue[ -1 ] != '\n':
                    starString += '\n'
                i = i + 1
            return starString

        ## Loop tags here
        starString += loop_ident_size * ' ' + 'loop_\n'

        for tagname in self.tagnames:
            ## Just format it such that it will take the least space
            starString += tagnames_ident_size * ' ' + '%s\n' % tagname
        starString += '\n'

        col_count = len( self.tagnames )
        row_count = len( self.tagvalues[0] )
        col_range = list(range( col_count))
        row_range = list(range( row_count))

        str_row = []
        row_id = 0
        tag_id = 0

        #
        # Wim modifications
        # Get formatting information sorted...
        # Would be much easier if I just set it as a float/integer/whatever!
        #

        colFormats = []
        colDataFuncs = []

        for tagvalues_column in self.tagvalues:
          isFloat = [0,0]
          isInteger = True
          maxLen = 0

          for tagvalue in tagvalues_column:

            if not tagvalue:
              continue

            if isFloat:
              floatSearch = self.floatPatt.search(tagvalue)
              if not floatSearch:
                isFloat = False
              else:
                numbers = len(floatSearch.group(1))
                decimals = len(floatSearch.group(2))
                if numbers > isFloat[0]:
                  isFloat[0] = numbers
                if decimals > isFloat[1]:
                  isFloat[1] = decimals

            if isInteger and not self.integerPatt.search(tagvalue):
              isInteger = False

            if maxLen < len(tagvalue):
              maxLen = len(tagvalue)

          if isInteger:
            colFormats.append(self.colspacer + "%" + str(maxLen) + "s")
            colDataFuncs.append(None)
          elif isFloat:
            colFormats.append(self.colspacer + "%" + str(isFloat[0] + isFloat[1] + 1) + "." + str(isFloat[1]) + "f")
            colDataFuncs.append(float)
          else:
            colFormats.append(self.colspacer + "%-" + str(maxLen + 2) + "s")
            colDataFuncs.append(None)


        ## This will quicken further actions and in itself is rather
        ## quick as we're using build in functions. I need to do this because
        ## I don't know of any splicing method that can get me a row from
        ## the table. The command below clocked 0.2 sec. on 0.6 Mb table (including
        ## spaces before parsing it was 1 Mb)
        ## Any quicker method in other modules?
        tagvalues_tr = transpose( self.tagvalues )

        for row_id in row_range:

            str_tmp = str.join( tagvalues_tr[row_id], ',' )

            ## Are quotes needed? Do it per row first to get some speed perhaps
            str_tmp = ''
            for col_id in col_range:
                colFormat = colFormats[col_id]
                ## Just format it such that it will take the least space
                if pattern_quotes_needed.search( self.tagvalues[col_id][row_id] ):
                    str_tmp = str_tmp + colFormat % quotes_add( self.tagvalues[col_id][row_id] )
                else:
                    tagvalue = self.tagvalues[col_id][row_id]
                    if colDataFuncs[col_id]:
                        tagvalue = colDataFuncs[col_id](tagvalue)
                    str_tmp = str_tmp + colFormat % tagvalue

            str_row.append( str_tmp )

            ## Perhaps delete for speed later on...
            if self.verbosity > 1:
                tag_id = tag_id + col_count
                if tag_id - count > count_hash:
                    count = tag_id
                    print('##### %s looped tag values collected ######' % count_hash)

        if show_stop_tag:
            str_row.append( '\n' + loop_ident_size * ' ' + 'stop_\n' )

        starString += str.join( str_row, '\n')

        # Save some space
        del tagvalues_tr

        return starString

    """
    A title identifing a tagtable by its tagnames
    simply the space separated concatenation of the tag names
    """
    def set_title ( self ):
        if self.verbosity >= 9:
            print('Setting title of tagtable')
        self.title = str().join( self.tagnames )


    """
    Size and type checks to be extended
    0 Only fast checks
    9 Type checks of each element
    """
    def check_integrity( self, check_type=0 ):

        names_length    = len(self.tagnames)
        values_length   = len(self.tagvalues)

        if names_length != values_length:
            print("ERROR: names_length[%s] != values_length[%s]:" % (
                names_length, values_length ))
            print("ERROR: names:", self.tagnames)
            return 1

        column_length_first = len( self.tagvalues[ 0 ] )
        for tag_id in range( values_length ):
            if len( self.tagvalues[ tag_id ] ) != column_length_first:
                print("ERROR: length column[%s](%s) is not the same as" % (
                            self.tagnames[ tag_id],
                            len( self.tagvalues[ tag_id ] ) ))
                print("ERROR: length column[%s](%s)" % (
                            self.tagnames[ 0],
                            column_length_first ))
                return 1

        if check_type >= 9:
            cols = list(range( names_length))
            rows = list(range( column_length_first))
            for row_id in rows:
                for col_id in cols:
                    val_type = type(self.tagvalues[col_id][row_id])
                    if val_type !=  bytes:
                        print("ERROR: type %s is not allowed as a value in a tagtable" % val_type)
                        print("ERROR: found for tagtable[%s][%s]" % ( self.tagnames[ col_id ], row_id ))
                        return 1

        if self.verbosity >= 9:
            print('Checked integrity of TagTable (%2s names %4s values each): OK [%s]' % (
                names_length, column_length_first, self.title ))
        return 0


    """
    - Parses text into a tagtable.
    - Returns the position in the string with the first non-white space
    character after the tagtable or the length of the text in case all
    was parsed. Just to be verbose, if the tagtable is ended by a save_
    then the starting position of the save_ will be returned.
    - Assumptions here is that ;; blocks are collapsed, see Text functions
    - For speed purposes I scan ahead to see how far I can go before
    hitting a quoted tag value. I estimate in the large tables only 1 in
    1000 has a ;; block and only 1 in 5-10 has '' or "" block. For the part
    that is not quoted the parsing can be really fast.
    """
    def parse(  self,
                text      = '',
                pos       = 0 ):
        ## Parse free tagtable reading all tag name/value pairs
        if self.free:
            pos = self._tagtable_free_parse( text, pos )
            if pos == None:
                print("ERROR: tagtable_free_parse returned with ERROR")
                return None
            else:
                return pos
            if self.check_integrity():
                print("ERROR: integrity of parsed table is not ok")
                return None

        ## Parse looped tagtable
        # Tag names
        match_tags_loop = pattern_tags_loop.search(text, pos)
        if not match_tags_loop:
            print("ERROR: No tag names found for looped tagtable")
            return None

        ## Do a limited search with findall for tag names
        match_tags_loop_2 = pattern_tags_loop_2.findall(text,
                                pos,
                                match_tags_loop.end() )
        for m in match_tags_loop_2:
            self.tagnames.append( m )
        pos = match_tags_loop.end()

        # End of loop
        ## There is no escaping these expensive searches if we can't depend
        ## on a stop sign
        ## Can be optimized further... by looking only up to the
        ## position already know to have a stop sign. Problem is that this
        ## is different for NMR-STAR (\sstop_) and mmCIF (\sloop_ or \s_\S)
        ## The (\ssave_) is included for when there are more flavors...

        text_length = len(text)
        if pos == text_length:
            print("ERROR: No tag values found for looped tagtable")
            return None

##        pos_sf_begin_or_end_nws = pattern_unquoted_find(text, pattern_sf_begin_or_end, pos)
        pos_tagtable_loop = pattern_unquoted_find(text, pattern_tagtable_loop_2, pos)
        pos_tagtable_stop = pattern_unquoted_find(text, pattern_tagtable_stop_2, pos)
        pos_tagname       = pattern_unquoted_find(text, pattern_tagname_2, pos)

        ## Find the first one and set the end postion to the beginning of
        ## the match excluding the beginning white space character
        pos_end = text_length
        if pos_tagtable_loop != -1 and pos_tagtable_loop<pos_end:
            pos_end = pos_tagtable_loop + 1
        if pos_tagtable_stop != -1 and pos_tagtable_stop<pos_end:
            pos_end = pos_tagtable_stop + 1
        if pos_tagname != -1 and pos_tagname<pos_end:
            pos_end = pos_tagname + 1

        if self.verbosity >= 9:
            print('pos_tagtable_loop:', pos_tagtable_loop)
            print('pos_tagtable_stop:', pos_tagtable_stop)
            print('pos_tagname      :', pos_tagname)
            print('Will parse tagtable text to end at position: [%s]' % pos_end)

        ## Just checking
        if not ( pos_tagtable_loop!=-1 or pos_tagtable_stop!=-1 or pos_tagname!=-1 ):
            if self.verbosity > 1:
                print('WARNING: EOF in tagtable, must be a CIF file')
##                print 'Items looked for are a begin or end of a saveframe, or'
##                print 'a begin (loop_) or end (stop_ or _tagname) of a tagtable'
##                print '(free or looped).'
##                print 'Actually the begin/end of saveframe is not checked since'
##                print 'NMR-STAR and mmCIF both end a tagtable without it.'

        # Tag values
        if self._tagtable_loop_values_parse(
                text, pos, pos_end): ## will set title too
            print("ERROR: not parsed table")
            return None
        ## Set the position to the end of this tagtable at the beginning
        ## of a stop_ or a new tagtable
        pos = pos_end

        ## Skip the stop sign and empty space if it was stop_
        if pos_tagtable_stop != -1:
            ## Try a match from the previously found position including
            ## the white space char before it.
            match_tagtable_stop = pattern_tagtable_stop_2.search( text, pos-1 )
            if not match_tagtable_stop:
                print("ERROR: no stop_ on second try")
                return None
            pos = match_tagtable_stop.end()

        if self.check_integrity():
            print("ERROR: integrity of parsed table is not ok")
            return None
        return pos


    """
    Parse names and values of free tagtable loop from pos
    returns new position alias status (None for failure)
    """
    def _tagtable_free_parse( self, text, pos ):

        text_length = len(text)

        while pos < text_length - 1:
            if text[pos] != '_':
                break
            # Tag name
            match_tag_name = pattern_tag_name.search(text, pos)
            if match_tag_name:
                if ( match_tag_name.start() - pos ) != 0:
                    print("ERROR: looking for a free tag name (0)")
                    return None
            else:
                print("ERROR: looking for a free tag name(1)")
                return None
            self.tagnames.append( match_tag_name.group(1) )
            pos  = match_tag_name.end()
            # Tag value
            value, pos = tag_value_parse(text, pos)
            if pos == 0:
                print("ERROR: looking for a free tag name(1)")
                return None
            ## Structures of free and looped tagtable are the same
            self.tagvalues.append( [ value ] )
            if self.verbosity >= 9:
                print('**Parsed tag name : [%s] and value [%s]: ' % (
                    match_tag_name.group(1), value))
        self.set_title()
        return pos


    """
    Parse values of tagtable loop from pos to pos_end
    returns status (None for success, 1 for failure)
    """
    def _tagtable_loop_values_parse( self, text, pos, pos_end):

        if self.free:
            print("ERROR: This is a 'free' tagtable, only looped tagtable can be parsed")
            return 1

        names_length                = len(self.tagnames)

        ## Empty the table
        self.tagvalues   = []
        for _tag_id in range( names_length ):
            self.tagvalues.append( [] )

        ## Get rid of initial white space if any, shouldn't be needed
        match_white_space = re.compile('\s+').search( text, pos, pos_end )
        if match_white_space:
            if match_white_space.start() == 0: # Match has to start at the beginning
                pos = match_white_space.end()

        tag_id          = 0
        count           = 0
        count_hash      = 100000
        text_length        = len(text)

        ## Only process characters to predetermined end
        while pos < pos_end:
            if self.verbosity > 1:
                if pos - count > count_hash:
                    print('##### %s chars processed ######' % count_hash)
                    count = pos
            ## 1 char search; ', ", or (; at beginning of line)
            match_quoted = pattern_quoted.search( text, pos, pos_end )

            #
            # Wim 25/09/03: Changed following to allow correct parsing of H5'' type names
            # and "asdfasdf'" type stuff
            #

            if match_quoted:

                if match_quoted.start() == pos:
                    ## QUOTED:
                    value, pos = tag_value_quoted_parse( text, pos )

                    if pos ==  None:
                        print('ERROR: got error in parse (1)')
                        return 1
                    if pos > pos_end:
                        print('ERROR: found a quoted value that was not wholly within boundaries (1)')
                        return 1
                    self.tagvalues[ tag_id ].append( value ); tag_id = tag_id + 1

                    if tag_id == names_length:  tag_id = 0

                else:

                    #
                    # New positions depend on whether correct quote or not
                    # If not correct quote, reset pos and do 'normal' parse
                    #

                    if text[match_quoted.start()] in ["'",'"'] and text[match_quoted.start()-1] != " ":

                        tempendpos = match_quoted.start() + str.find(text[match_quoted.start():],' ')

                    else:

                        tempendpos = match_quoted.start()

                    ## Parse all unquoted tag values beginning from position
                    ## UP TO specified end position
                    ## NOT QUOTED

                    for t in text[pos:tempendpos].split():
                        self.tagvalues[tag_id].append( t ); tag_id = tag_id + 1
                        if tag_id == names_length: tag_id = 0

                    if tempendpos == match_quoted.start():

                        ## QUOTED:
                        pos = tempendpos

                        value, pos = tag_value_quoted_parse( text, pos )

                        if pos ==  None:
                            print('ERROR: got error in parse (2)')
                            return 1
                        if pos > pos_end:
                            print('ERROR: found a quoted value that was not wholly within boundaries (2)')
                            return 1
                        self.tagvalues[ tag_id ].append( value ); tag_id = tag_id + 1
                        if tag_id == names_length:  tag_id = 0

                    else:

                        pos = tempendpos

            else:

                ## NOT QUOTED until end (only executed once)
                for t in text[pos:pos_end].split():
                    self.tagvalues[tag_id].append( t ); tag_id = tag_id + 1
                    if tag_id == names_length: tag_id = 0

                pos = text_length # Needed to break while loop


        col_length = len( self.tagvalues[-1] )

        if tag_id != 0:
            print("ERROR: not correct number of tag values read")
            print("Read [%s] tag(s) that is:" \
                  % ( col_length * names_length + tag_id ))
            print("[%s] row(s) complete and [%s] tag value(s) in last row that is incomplete." \
                  % ( col_length, tag_id ))
            print("Tag names of this table are:")
            print(self.tagnames)

            for xxx in range(0,len(self.tagvalues[0])):

              for yyy in range(0,len(self.tagvalues)):
                print(self.tagvalues[yyy][xxx], end=' ')

              print('\n-----------------------------------------------')

            pos = 0
            while pos < tag_id:
                 print(self.tagvalues[pos][-1])
                 pos = pos + 1

            return 1

        if col_length == 0:
            print("ERROR: no tag values parsed")
            return 1

        # Set the title
        self.set_title()
        return None


###############################################################################
if __name__ == '__main__':
    STAR.verbosity = 9
##    text = """_A a _B "b'" """; free = 1
    text = """_A a b c"""; free = None
    tt = TagTable(  free      = free,
                    tagnames  = [],
                    tagvalues = [],
                    verbosity = STAR.verbosity)
    pos = tt.parse( text = text, pos = 0)
    print('got pos: [%s] for length: [%s] (Should be the same)' % (pos, len(text)))
    print(tt)
    print(tt.star_text())



"""
Just a few utilities that can be of more general use.
"""

__author__    = "$Author: wimvranken $"
___revision__ = "$Revision: 1.4 $"
___date__     = "$Date: 2007-07-09 12:22:27 $"

"""
$Log: not supported by cvs2svn $
Revision 1.3.8.1  2007/06/13 08:35:52  wimvranken
Removed copyright - is not CCPN

Revision 1.3  2005/06/27 16:43:42  wb104
Updated licenses.

Revision 1.2  2004/12/15 17:57:20  tjs23
TJS: Updated licenses.

Revision 1.1  2003/07/01 12:56:19  wfv20
Jurgen Doreleijers modified Python nmrStar reader. Bug fixes and additions (notably for reading nmrView star files) have been made.

Revision 1.1.1.1  2001/11/02 20:16:40  jurgen
Initial package capable of read/write access to STAR files without nested loops

"""

class Lister:
    """Example from 'Learning Python from O'Reilly publisher'"""
    def __repr__(self):
        return ("<Instance of %s, address %s:\n%s>" %
           (self.__class__.__name__, id(self), self.attrnames()))

    def attrnames(self):
        result=''
        keys = list(self.__dict__.keys())
        keys.sort()
        for attr in keys:
            if attr[:2] == "__":
                result = result + "\tname %s=<built-in>\n" % attr
            else:
                result = result + "\tname %s=%s\n" % (attr, self.__dict__[attr])
        return result        


"""
A fast transposing algorithm from the python mailing list
"""
def transpose ( matrix ):
    if len( matrix ) < 1:
        print('ERROR: trying to transpose an empty matrix')
        return 1
    elif len( matrix ) == 1:
        if len(matrix[0]) == 0:
            print('ERROR: trying to transpose an empty matrix, shape would be lost')
            print('ERROR: [[]] would become []')
            return 1
        else:
            return [(y,) for y in matrix[0]]
    else:
        return list(map(*[None,] + list(matrix)))

###############################################################################

if __name__ == '__main__':
    if 1:
        m = [ [1,2], [3,4] ]
        print(m, 'is transposed:', transpose(m))



import pickle, os

from b2bTools.singleSeq.DynaMine.Predictor import DynaMine
from b2bTools.general.Io import B2bIo

class EFoldMine(B2bIo):

  version = '2.0'
  # scriptName = "b2bTools.singleSeq.EFoldMine.Predictor"
  scriptName = "python.b2bTools.singleSeq.EFoldMine.Predictor"
  name = "EFoldMine"

  # pickleProbabilisticModelFile = os.path.abspath(os.path.join(os.path.dirname(__file__),'models',"efModelRBF2016.proba.cPickle"))
  pickleProbabilisticModelFile = os.path.abspath(os.path.join(os.path.dirname(__file__),'models',"efoldmine_converted.cPickle"))

  discreteCutoff = 0.196 # or 0.192?

  def __init__(self):
    self.allPredictions = {}
    self.window = 2

    print("Reading the early folding probabilistic model...")
    self.model = pickle.load(open(self.pickleProbabilisticModelFile,'rb'),encoding='latin1')

    print("Initializing Dynamine...", end=' ')
    self.dynaMine = DynaMine(predictionTypes=('backbone', 'sidechain', 'coil', 'sheet', 'helix'))
    print("Done.")

    # Additional info for writing files
    self.references = ['doi: 10.1038/ncomms3741 (2013)', 'doi: 10.1093/nar/gku270 (2014)', 'doi: 10.1038/s41598-017-08366-3 (2017)']
    self.infoTexts = ['Generated by EFoldMine','Based on the DynaMine project','See http://bio2byte.be']

    self.informationPerPredictor = {
      self.name: {
          'references': self.references,
          'info':    ";".join(self.infoTexts),
          'version': self.version,
          'origin':  self.scriptName
                }
          }

  def predictSeqs(self,seqs,dynaMinePreds=None, includeDiscreteClass=False):

    """
    :param seqs: List/tuple of (seqId,sequenceString) tuples
    :param dynaMinePreds: DynaMine predictions matching info in seqs
          in self.allPredictions[seqId][predictionType] = [(aminoAcidTypeString,predValue),...] format
    :keyword includeDiscreteClass: If set to True, prediction output will include a class prediction
             (1 for early folding, 0 for not early folding)
    :return: True if all sequences predicted, False if problems

    See self.allPredictions for the predictions themselves. Will include DynaMine values.
    """

    print("Start predictions...")

    # These can be fed in if available, if not then will run these automatically
    if not dynaMinePreds:

      #
      # DynaMine predictions if necessary
      #

      self.dynaMine.predictSeqs(seqs)
      self.allPredictions = self.dynaMine.allPredictions

    else:
      self.allPredictions = dynaMinePreds

    # Now do early folding predictions
    seqDoneCount = 0
    for (seqId,sequence) in seqs:

      if len(sequence) < 7:
        print(("Sequence with ID {} too short, ignoring".format(seqId)))
        continue

      #
      # Separate variable where the value of the DynaMine backbone preds is 1 - necessary for the early folding prediction!
      #

      dynaMineBackboneNormalised = self.allPredictions[seqId]['backbone'][:] # Make a copy, will be shifted to maximum value of 1, might mess up other code

      maxValue = max([values[1] for values in dynaMineBackboneNormalised])
      correction = 1 - maxValue
      for seqIndex in range(len(dynaMineBackboneNormalised)):
        dynaMineBackboneNormalised[seqIndex] = (dynaMineBackboneNormalised[seqIndex][0],dynaMineBackboneNormalised[seqIndex][1] + correction)

      #
      # Run the early folding prediction
      #

      x = self.buildVectors((sequence,
                             dynaMineBackboneNormalised,
                             self.allPredictions[seqId]["helix"],
                             self.allPredictions[seqId]["sheet"],
                             self.allPredictions[seqId]["coil"],
                             self.allPredictions[seqId]["sidechain"]),
                             self.window)

      yp = self.model.predict_proba(x)[:,1]

      #
      # Add to prediction variable
      #

      assert len(yp) == len(sequence)

      # Adding to DynaMine predictions for consistency
      self.allPredictions[seqId]['earlyFolding'] = []

      for i in range(len(sequence)):
        if includeDiscreteClass:
          predTuple = (sequence[i], yp[i], self.discretePreds(yp[i]))
        else:
          predTuple = (sequence[i], yp[i])

        self.allPredictions[seqId]['earlyFolding'].append(predTuple)

      seqDoneCount += 1

    allDone = False
    if len(seqs) == seqDoneCount:
      allDone = True

    return allDone

  def discretePreds(self,v):

    if v > self.discreteCutoff:
      return 1
    return 0

  def buildVectors(self,feats, window):
    x = []
    y = []
     #          			 0		1		2		3	4		5		  6	  7		8		9			10
    tmp = feats#db[name] = (seq, backbone, helix, sheet, coil, sidechain, s2, rsa, ground, espritzNmr, espritzXray)
    i = 0
    while i < len(tmp[0]):
      #tx, ty = getSingleFeats(tmp, i)
      # 1 backbone, 2 helix,  3 sheet,  4 coil, 5 sidechain
      tx = self.getWindowFeats(tmp, 1, i, window)+ self.getWindowFeats(tmp, 2, i, window)+self.getWindowFeats(tmp, 3, i, window)+self.getWindowFeats(tmp, 4, i, window)+self.getWindowFeats(tmp, 5, i, window)
      assert len(tx) == 25
      x.append(tx)
      i += 1
    return x

  def getWindowFeats(self,data, featureNum, pos, w):	#feature indicates : 1 backbone, 2 helix,  3 sheet,  4 coil, 5 sidechain
    # 1 backbone, 2 helix,  3 sheet,  4 coil, 5 sidechain
    chosenFeat = data[featureNum][max(pos-w, 0):min(pos+w+1,len(data[1]))]
    chosenFeatW = []
    if pos-w < 0:
      chosenFeatW = [-1] * -(pos-w)
    for i in chosenFeat:
      chosenFeatW.append(round(i[1],3))
    if pos+w+1 > len(data[featureNum]):
      chosenFeatW += [-1] * (pos+w+1 - len(data[featureNum]))
    assert len(chosenFeatW) == (w*2 +1)
    #print chosenFeatW
    return chosenFeatW



# -*- coding: utf-8 -*-
import sysconfig, os
from b2bTools.singleSeq.PSPer.phase_transition_hmm import phase_hmm
import string
import numpy as np


def leggifasta(database):
		f=open(database)
		uniprot=f.readlines()
		f.close()
		dizio={}
		for i in uniprot:
			if i[0]=='>':
					uniprotid=i.strip('>\n')
					dizio[uniprotid]=''
			else:
				dizio[uniprotid]=dizio[uniprotid]+i.strip('\n').upper()
		return dizio

def standalone(input_obj):
	features='dyna_back,psipred,ef,dyna_side'
	window=4
	verbose=0

	def check_sequences(seqs):
		for i in list(seqs.keys()):
			if i=='extra_predictions':
				continue
			if not seqs[i].isalpha():
				return {'error':'invalid char in sequence '+i}
			if len(seqs[i])>3000:
				#print len(seqs[i])
				return {'error':'sequence '+i+' too long, maximum length is 3000 amino acids'}
			if len(seqs[i])<20:
				#print len(seqs[i])
				return {'error':'sequence '+i+' short, minimum length is 20 amino acids'}

		return True

	def load_model():
		mod = phase_hmm()
		mod.fit()

		scaler = None

		return mod, scaler

	def format_output(disorder, viterbi, seqs, features):
		protein_id_keys = list(disorder.keys())

		out=[]
		for protein_id_index in range(len(protein_id_keys)):
			protein_id = protein_id_keys[protein_id_index]
			current_features = features[protein_id]

			for feature_index in range(5):
				actual = current_features[:, feature_index]
				base   = current_features[:, 0]

				assert len(actual) == len(base)

			entry={}

			features[protein_id] = np.array(current_features)
			entry['proteinID'] = protein_id
			entry['sequence'] = seqs[protein_id]
			entry['protein_score']=disorder[protein_id]
			entry['viterbi']=viterbi[protein_id] if protein_id in viterbi else np.full(len(seqs[protein_id]), '')
			entry['complexity']=list(features[protein_id][:,0])
			entry['arg']=list(features[protein_id][:,1])
			entry['tyr']=list(features[protein_id][:,2])
			entry['RRM']=list(features[protein_id][:,3])
			entry['disorder']=list(features[protein_id][:,4])

			out+=[entry]

		return { 'results': out }

	def predict_fasta(sequence_input, model, crunch=100):
		if type(sequence_input)==str:
			try:
				fasta_sequences=leggifasta(sequence_input)
			except:
				return {'error':"problems in the fasta file"}
		elif type(sequence_input)==dict:
			fasta_sequences=sequence_input
		else:
			return {'error':'internal error, wrong object passed to the standalone, it must be either a dict or a string'}

		check = check_sequences(fasta_sequences)
		if not check:
			return check

		# targets=list(fasta_sequences.keys())[:]
		# results_dict={}
		# cont=0
		# dyna={}
		# side={}
		# ef={}

		built_vector = model.build_vector(fasta_sequences)
		results_dict = model.predict_proba(built_vector)
		viterbi = model.viterbi(built_vector)

		dict_features = {}
		printable_characters = string.printable
		for id in built_vector.keys():
			seq_vector = []

			for seq_vector_element in built_vector[id]:
				temp_seq_vector = []
				# feature_counter = 0

				for element_feature in seq_vector_element:
					feature_char_index = printable_characters.index(element_feature)
					temp_seq_vector += [float(feature_char_index)]

					# if feature_counter == 1 or feature_counter == 2:
					# 	feature_char_index = printable_characters.index(element_feature)
					# 	temp_seq_vector += [float(feature_char_index)]
					# else:
					# 	feature_char_index = printable_characters.index(element_feature)
					# 	temp_seq_vector += [float(feature_char_index)]

					# 	feature_counter += 1

				seq_vector += [temp_seq_vector]

			dict_features[id] = np.array(seq_vector)

		results = format_output(results_dict, viterbi, fasta_sequences, dict_features)
		return results

	# This should not be necessary with new disomine setup, commented out
	#clean_psipred_tmp()
	phase_hmm_instance, _ = load_model()
	results = predict_fasta(input_obj, phase_hmm_instance)

	return results

def main(args):
	#print standalone("example.fasta")
	fasta_sequences=leggifasta('input_files_examples/example_toy.fasta')
	fasta_sequences['extra_predictions']=False
	print(standalone(fasta_sequences))
	#from memory_profiler import memory_usage
	#mem_usage = memory_usage(standalone,interval=0.01)
	#print('Memory usage (in chunks of .1 seconds): %s' % mem_usage)
	#print('Maximum memory usage: %s' % max(mem_usage))
	#cProfile.run('standalone("example.fasta")')

if __name__ == '__main__':
	import sys
	sys.exit(main(sys.argv))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  chemical_shifts_custom_model.py
#  
#  Copyright 2019  
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
letters={'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K', 'ASN': 'N', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ALA': 'A', 'HIS': 'H', 'GLY': 'G', 'ILE': 'I', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}


shifts={ 'A': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'C': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'E': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'D': ['HA', 'CA', 'CB', 'C', 'HB2', 'N', 'H'],
		 'G': ['HA3', 'HA2', 'CA', 'C', 'N', 'H'],
		 'F': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'I': ['HA', 'CA', 'CB', 'C', 'HD11', 'HB', 'N', 'H'],
		 'H': ['HA', 'CA', 'CB', 'C', 'N', 'H', 'HB2', 'HD2'],
		 'K': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'M': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'L': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N', 'HD11', 'HD21'],
		 'N': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'Q': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'P': ['HA', 'CA', 'CB', 'C', 'HG2', 'HB2', 'CD', 'HD2'],
		 'S': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'R': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'T': ['HA', 'CA', 'CB', 'C', 'HG21', 'HB', 'N', 'H'],
		 'W': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N'],
		 'V': ['HA', 'CA', 'CB', 'C', 'HG11', 'HG22', 'N', 'H'],
		 'Y': ['HA', 'CA', 'CB', 'C', 'HB2', 'H', 'N']
 }




"""
Classes for dealing with STAR syntax
"""

__author__    = "$Author: jurgenfd $"
___revision__ = "$Revision: 1.12.4.2 $"
___date__     = "$Date: 2011-04-06 09:00:38 $"

"""
$Log: not supported by cvs2svn $
Revision 1.12.4.1  2010/02/03 09:32:06  wimvranken
Will now quote string if it containts a ,

Revision 1.12  2007/10/12 14:14:41  wimvranken
Took out fix to handle semicolons - it is valid NMR-STAR to have other chars after this one.

Revision 1.11  2007/07/09 12:22:27  wimvranken
Merge with branch 4.

Revision 1.10.4.1.4.1  2007/06/13 08:36:19  wimvranken
Removed copyright - is not CCPN
Bug fix for final _save tag

Revision 1.10.4.1  2006/05/23 16:41:41  wimvranken
Bug fix by Jurgen Doreleijers (BMRB) in semicolons_add()

Revision 1.10  2005/12/19 15:00:26  wimvranken
Fixed other small bug in ; comment handling

Revision 1.9  2005/12/15 18:19:34  wimvranken
Removed some previous changes for ; handling

Revision 1.8  2005/11/03 14:08:23  wimvranken
Slightly better handling of ; text blocks

Revision 1.7  2005/06/27 16:43:42  wb104
Updated licenses.

Revision 1.6  2004/12/15 17:57:20  tjs23
TJS: Updated licenses.

Revision 1.5  2004/06/29 11:40:57  wfv20
Better handling of writing strings with # in them

Revision 1.4  2003/10/29 11:10:12  wfv20
Now handles quotation marks within quotation marks if not followed by white space

Revision 1.3  2003/08/08 12:45:39  wfv20
Changed preferred quote

Revision 1.2  2003/07/24 15:28:33  wfv20
Better handling of hashes in quotes

Revision 1.1  2003/07/01 12:56:19  wfv20
Jurgen Doreleijers modified Python nmrStar reader. Bug fixes and additions (notably for reading nmrView star files) have been made.

RRevision 1.1.1.1  2001/11/02 20:16:40  jurgen
IInitial package capable of read/write access to STAR files without nested loops

"""
## Standard modules
import string, re

## BMRB modules
from b2bTools.general.bmrb import base as STAR

"""
Some handy patterns and functions for dealing with text in the STAR syntax.
Some are complicated because in Python the none-greedy pattern matching
gets too recursive and will actually bomb on larger strings. Like the
following code causes a bomb:
re.search( 'a.*?c', 'a' + 99999*'b' + 'c' )
Produces: 'RuntimeError: maximum recursion limit exceeded'
"""

## Since there are only functions and no classes in this module
## the verbosity may be changed by changing the variable directly.
## I know that's vague but I don't know how to do it yet... todo.
verbosity           = 2

## When not sure if text can have a ; at start of line use
## this string prepended to each line.
prepending_string   = '[raw] '


## Following string will be replacing the eol in a semicolon block where needed
## It may not contain any funny characters and shouldn't have underscores
## because it will make parsing slower. Parentheses, if used, should be of the
## square type.
eol_string = '<eol-string>'
eol_string_length = len(eol_string)

pattern_tagtable_loop = re.compile(r"""
^\s*  loop_  \s*                                  # Begin of loop
(   ^\s*     (?P<tagname>_\S+) \s*\n  )+          # Tag names with some spaces
       (?P<rawtext>.+?)                           # Tag table raw text
^\s*  stop_  \s*\n                                # End of loop
     """, re.DOTALL | re.MULTILINE | re.VERBOSE )

pattern_semicolon_block = re.compile(r"""
    ^;                                          # semicolon at begin, any text and then eol
    .+?                                         # Raw text for match object but not greedy
    ^;                                          # semicolon at begin, that's it
     """, re.DOTALL | re.MULTILINE | re.VERBOSE )

pattern_eol_string     = re.compile( eol_string, re.MULTILINE )

## Next pattern tells when search for on ONE tagvalue if it needs quotes
pattern_quotes_needed  = re.compile( r'[\s\'\"\,]|^_|^\#' )

## Next pattern tells when search for on MANY tagvalues if it needs quotes
## The values should be joined by a comma. A value: 'bla,_bla' will be
## mentioned as needing quotes unnecessarily but that's dealt with in the code by further checking
pattern_quotes_needed_2= re.compile( r'[\s\'\"]|^_|,_|,\#' )

pattern_eoline_etcet   = re.compile( r'[\n\r\v\f]' )
pattern_single_qoute   = re.compile( r"'" )
pattern_double_qoute   = re.compile( r'"' )

pattern_save_begin      = re.compile('save_(\S+)\s+')
pattern_save_end        = re.compile('save_\s*')
pattern_tagtable_loop   = re.compile("loop_\s*" )
pattern_tagtable_stop   = re.compile("stop_\s*" )
# Same thing but not eating all white space chars, just a minimal match
pattern_save_begin_nws      = re.compile('save_\S')
#pattern_save_end_nws        = re.compile('save_\s')
pattern_save_end_nws        = re.compile('(?:save_\s)|(?:save_$)')
pattern_tagtable_loop_nws   = re.compile('loop_\s')
pattern_tag_name_nws        = re.compile('_\S')
# Same thing but requiring a prefixed white space char:
##pattern_sf_begin_or_end = re.compile('\ssave_')
pattern_tagtable_loop_2 = re.compile('\sloop_\s+' )
pattern_tagtable_stop_2 = re.compile('\sstop_\s+' )
pattern_tagname_2       = re.compile('\s_\S+\s+' )

pattern_tag_name = re.compile(r"""(_\S+) \s+
     """, re.DOTALL | re.MULTILINE | re.VERBOSE )
pattern_tags_loop       = re.compile(r"""(?: (_\S+) \s* )+
     """, re.MULTILINE | re.VERBOSE )
pattern_tags_loop_2     = re.compile(r"""    (_\S+) \s*
     """, re.MULTILINE | re.VERBOSE )

## Get any number of non-white space characters followed by any white space
pattern_word            = re.compile(r"""(\S+)\s*""", re.MULTILINE )

pattern_quoted = re.compile(r"""
        ['"] |                          # single or double quote
    (?: ^ ; )                           # semicolon at the beginning of a line
     """, re.MULTILINE | re.VERBOSE )

pattern_quoted_2 = re.compile(r"""(?: \b [\'\"] ) | (?: ^  \;     )""", re.MULTILINE | re.VERBOSE )

pattern_s_quote        = re.compile(r"""\'\s+""", re.MULTILINE )
pattern_d_quote        = re.compile(r"""\"\s+""", re.MULTILINE )
pattern_e_semicolon    = re.compile( eol_string + r"""\;\s*""", re.MULTILINE ) # Added \n for better parsing Wim 01/11/05

# Set beginning of line BEFORE whitespace - Wim 06/03/2003
pattern_comment_begin  = re.compile (r"""^\s*\#.*\n           # A string starting a line with a sharp
                                   """, re.MULTILINE | re.VERBOSE)

pattern_nmrView_compress_empty = re.compile(r""" \{(\s+)\}
                                             """, re.MULTILINE | re.VERBOSE)
pattern_nmrView_compress_questionmark = re.compile(r""" \{(\s+\?)\}
                                                    """, re.MULTILINE | re.VERBOSE)

pattern_comment_middle = re.compile (r""" (^[^;^\n]
                                          (?:([\'][^\']*\#[^\']*[\']|[\"][^\"]*\#[^\"]*[\"])|[^\#.])*? )  # Any string beginning a line other than with a semicolon and with no quotes in it
                                          (\s+\#.*)?    $	    # Any string ending a line and starting with a sharp
                                   """, re.MULTILINE | re.VERBOSE)

    # Hashes in quotes don't count!
    # (?:[\'\"][^\'^\".]*\#[^\'^\".]*[\'\"]|[^\#.])*? ) expression gets '<text>#<text>' blocks,
    # is now built into multiline search, seems to be working... (Wim 11/02)
    #
    # Changed \s* to \s+ - comments can only start with a ' ' before the '#' (Wim 05/03)
    # Removed . from [^\'^\".] in regular expression described above: more generic (Wim 05/03)

"""
Searches for a regular expression in text.
The text may not be STAR quoted and must have semicolon blocks collapsed
such that the semicolon starts at the beginning of the line.
Returns the start position of the match or -1 if it was not found or
None if there was an error.

The function will search the text from given position onwards
and checks the chars preceding (up to the line it's in) for quote style.

WARNINGS:
- Don't call it for a text that has no \n and at least 1 other
character in it before pos (not fully tested; perhaps possible).
- I have not put in extra checks because of needed speed.
- No requirements set on what follows the pattern.
"""

def pattern_unquoted_find(text, pattern, pos=0):
    while 1:
        match = pattern.search( text, pos)
        if not match:
            ## No match at all
            return -1

        pos = match.start()

        ## Is it the beginning of the string
        if pos == 0:
            return 0

        ## Is the first character matched an eol it self
        if text[pos]=='\n':
            if verbosity >= 9:
                print('Found pattern: [%s] at the beginning of a line' % pattern.pattern)
            return pos

        ## I hope the rfind is optimized to stroll backwards from pos
        pos_end_of_previous_line = text.rfind('\n', 0, pos)
        if pos_end_of_previous_line == -1:
            pos_end_of_previous_line = -1 ## Dangerous rewind?

        line = text[pos_end_of_previous_line+1:pos]
        # Some dummy value but continue with the test below.
        if line == '':
            line = ' '

        # Not the one
        if line[0] == ';':
            if verbosity > 1:
                print('WARNING: (1) found pattern: [%s] preceded by: [%s]' % (
                    pattern.pattern, line ))
            pos = pos + 1
            continue

        squoted = None
        dquoted = None
        for i in line:
            if      i == "'":
                if not dquoted:
                    squoted = not squoted
            elif    i == '"':
                if not squoted:
                    dquoted = not dquoted
        if squoted or dquoted:
##            if squoted and dquoted:
##                ## Should not be possible to occur, delete when confident
##                print "ERROR: code error, mixing of quote styles in line:"
##                print "ERROR: [%s]" % line
##                return None
            if verbosity > 1:
                print('WARNING: (2) found pattern: [%s] preceded by: [%s]' % (
                    pattern.pattern, line ))

            # Not the one
            pos = pos + 1
            continue

        return pos


"""
Parse one quoted tag value beginning from position: pos
Return the value and the position of the 'cursor' behind the
value for the first non white space char.
In case of error the position value of None will signal failure.
"""
def tag_value_quoted_parse( text, pos ):
##    print 'text: [%s]' % text
##    print 'pos:  [%s]' % pos
    if text[ pos ] == '"':
        match_d_quote = pattern_d_quote.search( text, pos+1)
        if not match_d_quote:
            print("ERROR: No matching double quote char found for double quote char at offset:", 0)
            print("ERROR: Next 70 chars are: [%s]" % text[ pos:pos+70 ])
            return None, None
    ##            if verbosity >= 9:
    ##                print "pos, span():", pos, match_d_quote.span()
    ##                print 'Found Q tag value: [%s]' % text[ pos+1:match_d_quote.start() ]
        return text[ pos+1:match_d_quote.start() ], match_d_quote.end()

    if text[ pos ] == "'":
        match_s_quote = pattern_s_quote.search( text, pos+1)
        if not match_s_quote:
            print("ERROR: No matching single quote char found for single quote char at offset:", 0)
            print("ERROR: Next 70 chars are: [%s]" % text[ pos:pos+70 ])
            return None, None
        value = text[ pos+1:match_s_quote.start() ]
    ##            if verbosity >= 9:
    ##                print "pos, span():", pos, match_s_quote.span()
    ##                print 'Found Q tag value: [%s]' % value
        return value, match_s_quote.end()

    ## Remove check for speed if you want
    ## This should always be true
    if text[ pos ] == ";":
        match_e_semicolon = pattern_e_semicolon.search( text, pos+1)
        if not match_e_semicolon:
            print("ERROR: No matching semicolon found for semicolon char at offset:", 0)
            print("ERROR: Next 70 chars are: [%s]" % text[ pos:pos+70 ])
            return None, None
    ##            print "pos, span():", pos, match_e_semicolon.span()
        ## Include the first eol and the eol before the semicolon
        value = text[ pos+1:match_e_semicolon.start()+eol_string_length ]
        ## Expansion relatively cheap here and harmless if unique string as defined in
        ## eol_string is indeed unique
        ## print 'Found Q (semicolon) tag value: unexpanded [%s]' % value
        ## print '-----------'
        ## print text[ match_e_semicolon.start()+eol_string_length : match_e_semicolon.start()+eol_string_length + 20]
        ## print '-----------'
        value = semicolon_block_expand( value )
        ## print 'Found Q (semicolon) tag value: expanded   [%s]' % value

        return value, match_e_semicolon.end()

    print("ERROR: Position in text:", pos)
    print("""ERROR: should contain a ', ", or a ; but was not found:""")
    print("ERROR: Next 70 chars are: [%s]" % text[ pos:pos+70 ])
    return None, None


"""
From text on position pos, read a tag value and return the value and
position of the next non-space char. This is the slow parsing method
that should only be used for free tags.
"""
def tag_value_parse( text, pos):

    match_quoted = pattern_quoted.search( text, pos )
    if match_quoted:
        if match_quoted.start() == pos:

            return tag_value_quoted_parse( text, pos ) # Better speed with this code

    match_word = pattern_word.search( text, pos )
    if not match_word:
        print("ERROR: No match for a 'word' at offset:", pos)
        print("ERROR: Next 70 chars are:", text[ pos:pos+70 ])
        return None, None
    if match_word.start() != pos:
        print("ERROR: Match for a 'word' at wrong offset:", match_word.start() - pos)
        print("ERROR: Next 70 chars are:", text[ pos:pos+70 ])
        return None, None

    ## Include the first eol and the eol before the semicolon
    return  match_word.group(1), match_word.end()



"""
See function semicolon_block_collapse that calls this one
"""
def semicolon_block_replace( matchobj ):
    #print len(matchobj.group())
    return re.sub(  '\n', eol_string, matchobj.group() )


"""
This function should be called (not semicolon_block_replace)
Putting all semicolon separated values on one line
by replacing the eol within with a unique key value
that is to be remove later on by it's sibling method:
semicolon_block_expand.
SPEED:  0.6 cpu seconds for a 5 Mb file with 31 blocks and
        1.3 "                10 "            64 ".
"""
def semicolon_block_collapse( text ):

    count = 0
    startpos = 0

    # TODO: this is not good - since text[startpos:] is used it's always the start of a line, so if string starts with ;...
    pattern_semicolon_only = re.compile("^\;", re.MULTILINE)
    # Added special _end pattern with $ for better pattern matching - Wim 31/10/2005
    # WARNING: took out this fix because it is valid NMR-STAR to have other info after it
    # This will probably create problems with some restraint grid files...
    # If there is a text line starting with ;, it should get an empty space in front
    #pattern_semicolon_only_end = re.compile("(^\;\s*$)", re.MULTILINE)

    semicolon_start = pattern_semicolon_only.search(text[startpos:])

    while(semicolon_start):

      count += 1

      startpos = startpos + semicolon_start.start()
      semicolon_end = pattern_semicolon_only.search(text[startpos+1:])
      try:
        endpos = startpos + 1 + semicolon_end.end() #- len(semicolon_end.group(1)) + 1
      except:
        print(text[startpos:startpos+100])
        raise

      # FOR TESTING
      #print "@@@@"
      #print text[startpos:endpos]
      #print "@@@@"

      text_replace = re.sub("\n", eol_string,text[startpos:endpos])

      # This is bulky and not very elegant but works
      text= text[0:startpos] + text_replace + text[endpos:]

      startpos = startpos + len(text_replace)

      semicolon_start = pattern_semicolon_only.search(text[startpos:])

    # Original code: can't handle re matches that are too long
    #text, count = pattern_semicolon_block.subn( semicolon_block_replace, text )
    if STAR.verbosity >= 9:
        print('Done [%s] subs with semicolon blocks' % count)
    return text

def semicolon_block_expand( text ):
    return pattern_eol_string.sub('\n', text )

"""
Adds semicolons, single quotes or double quotes depending on
need according to star syntax.
Does not assume that no quotes exist initially and will strip them if
present in pairs only.

If the possible_bad_char parameter is set (to 1 or higher) then
strings that would normally end up in a semicolon delimited blob will
have a string inserted at the beginning to it. The string can be the 'p'
argument to this function. [TODO]
"""
def quotes_add( text ):

    preferred_quote='"' # This info should be in a more central spot

    if pattern_eoline_etcet.search( text ):
        return semicolons_add( text )

    if pattern_single_qoute.search( text ):
        single_qoute_match = 1
    else:
        single_qoute_match = 0

    if pattern_double_qoute.search( text ):
        double_qoute_match = 1
    else:
        double_qoute_match = 0

    if single_qoute_match and double_qoute_match:
        return semicolons_add( text )

    if single_qoute_match:
        return '"' + text + '"'
    # Commented out because it leads to the same behaviour
    if double_qoute_match:
        return "'" + text + "'"

    ## Space other than end of line, or # sign etc.
    return preferred_quote + text + preferred_quote


"Strips quotes in pairs and returns new/old string"
def quotes_strip( text ):

    ## Can it be containing quotes?
    if len(text) <= 1:
        return text
    for quote_symbol in [ "\'", '\"' ]:
        if ( text[0]  == quote_symbol and
             text[-1] == quote_symbol ):
            return text[1:-1]
    return text


"""
Returns the input with ; delimited, possibly with a string inserted at the beginning.
The string value should always be ended by a eol, otherwise
the second semicolon can not be the first char on a line.
"""
def semicolons_add( text, possible_bad_char=None ):
    if possible_bad_char:
        lines       = string.split(text,'\n')
        text   = ''
        for line in lines:
            text = text + prepending_string + line + '\n'
##        ## Code repeated for speed
##        return "\n;" + text + ";\n"
##    else:
##        return "\n;" + text + ";\n"
#JFD updates 5/23/2006; apparently the text does not always end with an eol.
    if not text.endswith('\n'):
       text = text + '\n'
    return "\n;" + text + ";\n"

"""
Strip the STAR comments
Timed at: 2.8 seconds for a 5 Mb file with 61 comments (comment begin=31/other=30)
Pentium 3 @ 800 MHz only counting cpu seconds
"""
def comments_strip( text ):
#    verbosity = 9
    text, _count = pattern_comment_begin.subn( '', text )
#    if verbosity >= 9:
#        print 'Done [%s] subs with comment at beginning of line' % count
    text, _count = pattern_comment_middle.subn( '\g<1>', text )
#    if verbosity >= 9:
#        print 'Done [%s] subs with comment not at beginning of line' % count
    return text

def nmrView_compress( text ):

    text, count = pattern_nmrView_compress_empty.subn( '{}', text )
    print('Compressed [%s] nmrView empty { } tags' % count)

    text, count = pattern_nmrView_compress_questionmark.subn( '{?}', text )
    print('Compressed [%s] nmrView question mark { ?} tags' % count)

    return text

###############################################################################
if __name__ == '__main__':
    if 0:
        text = """;
mmy xie
;;
__Test
"""
        print('got text:    [%s]' % text)
        text = semicolon_block_collapse( text )
        print('transformed: [%s]' % text)
        value, pos = tag_value_quoted_parse( text, 0 )
        print('parsed     : [%s]' % value)
        print('pos (after): [%s]' % pos)
    ##        print 'back:       [%s]' % t.semicolon_block_expand( text )

    if 1:
        text = """
; my text ## is no comment
## comment 1
value ## comment 2
"""
        print('Text (1):     [%s]' % text)
        print('Stripped: [%s]' % comments_strip( text ))




"""
======================COPYRIGHT/LICENSE START==========================

Constants.py: Useful constants for scripts in this directory and its subdirectories

Copyright (C) 2005 Wim Vranken (European Bioinformatics Institute)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.
 
A copy of this license can be found in ../../../../license/LGPL.license
 
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.
 
You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)
- PDBe website (http://www.ebi.ac.uk/pdbe/)

- contact Wim Vranken (wim@ebi.ac.uk)
=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Wim F. Vranken, Wayne Boucher, Tim J. Stevens, Rasmus
H. Fogh, Anne Pajon, Miguel Llinas, Eldon L. Ulrich, John L. Markley, John
Ionides and Ernest D. Laue (2005). The CCPN Data Model for NMR Spectroscopy:
Development of a Software Pipeline. Proteins 59, 687 - 696.

===========================REFERENCE END===============================
"""

#
# Translation between 1 letter code3Letter and '3 letter code'
# This '3 letter code' is used by Cns.
#

#nuclAcidCodes = {
bioPolymerCodes = {

     'RNA': [['A','C','G','U'],
             ['ADE','CYT','GUA','URA']],
     'DNA': [['A','DA','C','DC','G','DG','T','DT','U','DU'],
             ['ADE','ADE','CYT','CYT','GUA','GUA','THY','THY','URA','URA']],
     'protein': [['A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y'],
                 ['ALA','CYS','ASP','GLU','PHE','GLY','HIS','ILE','LYS','LEU','MET','ASN','PRO','GLN','ARG','SER','THR','VAL','TRP','TYR']]     
            }

nucleotideList = ['RNA','DNA']

defaultMolCode = ' '

defaultLowerDist = 1.8

defaultSeqInsertCode = ' '
defaultAltLoc = ' '

indent = "  "

atomOrder = ['A','B','G','D','E','Z','H']

chainCodeString = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

stableBondTypes = ('covalent','disulfide','link')
bondTypes = stableBondTypes + ('hydrogen','salt')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  hmmsearch_otf.py
#
#  Copyright 2018 scimmia <scimmia@scimmia-ThinkPad-L540>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  parseHmmer.py
#
#  Copyright 2018 Daniele Raimondi <daniele.raimondi@vub.be>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import os
import tempfile
from b2bTools.singleSeq.PSPer.Constants import hmmbuild_bin, hmmscan_bin

def parseBoundaries(line):
	# qwwe  42 IFVGQLDKETTREELNRRFSTHGKIQDINLIFK---PTNIFAFIKYETEEAAAAALESENHAIFLNKTMH 108
	tmp = line.strip().split()
	return tmp[1], tmp[-1]

def parsePosteriori(line):
	tmp = line.strip().split()
	r = []
	for i in tmp[0]:
		p = castPosterior(i)
		if p == -1:
			continue
		r.append(p)
	return r

def castPosterior(v):
	if v == "*":
		return 10
	elif v == ".":
		return -1
	return int(v)

def parseOut(f):
	with open(f) as ifp:
		lines = ifp.readlines()

	alscores = []
	domains = []
	length = -1
	l = 0

	while l < len(lines):

		if "Scores for complete sequences" in lines[l]:
			if "[No hits" in lines[l+5]:
				return None, None
			else:
				tmp = lines[l+4].strip().split()
				#print tmp
				bitscore = float(tmp[1])
				evalue = float(tmp[3])
				alscores.append(( bitscore, evalue))
		if "Domain annotation for each sequence" in lines[l]:
			while l < len(lines):
				if "== domain" in lines[l]:
					start, end = parseBoundaries(lines[l+3])
					domains.append((int(start), int(end), parsePosteriori(lines[l+4])))
					l+=4
				l+=1
				if "Internal pipeline statistics summary:" in lines[l]:
					break
		if "residues searched)" in lines[l]:
			tmp = lines[l].strip().split()
			#print tmp
			length = int(tmp[3][1:])

		l += 1

	return domains, length

def getFeatsFromHmmer(domains, length):

	if not domains:
		return None

	hmmer_features = [0] * length

	for domain in domains:
		start = domain[0]
		end = domain[1]
		j = start - 1

		while j < end:
			a = domain[2].pop(0)
			hmmer_features[j] = a
			j += 1

	return hmmer_features

def main():
	a, l =  parseOut("output")

	#print a, l
	#raw_input()
	b = getFeatsFromHmmer(a,l)
	#print b
	c=hmmer_wrapper()
	c.fit_hmmer()
	print (c.predict({'a':'AAAAAAAAAAAACCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA','B':'VYVGNLGNNGNKTELERAFGYYGPLRSVWVARNPPGFAFVEFEDPRDAADAVRELDGRTLCGCRVR'}))

class hmmer_wrapper:
	def __init__(self, root = ''):
		self.root = root
		self.ali_file = root + 'rrm_align.pfam'
		self.phase_trans_file = root + 'phase_trans.hmm'
		self.hmmbuild_bin = hmmbuild_bin
		self.hmmscan_bin = hmmscan_bin


	def fit_hmmer_command(self):
		return f'{self.hmmbuild_bin} {self.phase_trans_file} {self.ali_file} > /dev/null'


	def hmmscan_command(self, temp_hmmer_output_file, temp_fasta_file):
		return f'{self.hmmscan_bin} -o {temp_hmmer_output_file} {self.phase_trans_file} {temp_fasta_file}'


	def fit_hmmer(self):
		os.system(self.fit_hmmer_command())


	def predict(self, seqs):
		hmmer_results = {}

		for seq_key in seqs.keys():
			current_sequence = str(seqs[seq_key])

			temp_fasta_file = tempfile.NamedTemporaryFile(delete=False, prefix='tmp_', suffix='.fasta')
			temp_hmmer_output_file = tempfile.NamedTemporaryFile(delete=False, prefix='out_hmmer_', suffix='tmp')

			try:
				with open(temp_fasta_file.name, 'w') as f:
					f.write(f'>tmp\n{current_sequence}\n')

				os.system(self.hmmscan_command(temp_hmmer_output_file.name, temp_fasta_file.name))

				domains, length = parseOut(temp_hmmer_output_file.name)
				predicted_features_from_hmmer = getFeatsFromHmmer(domains, length)

				if not predicted_features_from_hmmer:
					predicted_features_from_hmmer = [0] * len(current_sequence)

				hmmer_results[seq_key] = predicted_features_from_hmmer
			finally:
				os.remove(temp_fasta_file.name)
				os.remove(temp_hmmer_output_file.name)

		return hmmer_results


if __name__ == '__main__':
	a=hmmer_wrapper()
	print (a.predict({'a':'AAAAAAAAAAAACCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA','B':'VYVGNLGNNGNKTELERAFGYYGPLRSVWVARNPPGFAFVEFEDPRDAADAVRELDGRTLCGCRVR'}))
